import com.opencsv.CSVReader;
import org.logicng.formulas.Variable;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.sql.SQLSyntaxErrorException;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.*;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public class Main {

    private static String maxsatSolverName = "unknown solver";
    private static NumberFormat numberFormat = new DecimalFormat("0");

    private static void writeFile(String resultDirectoryName, String fileName, StringBuilder stringBuilder) {
        try {
            File dir = new File(resultDirectoryName);
            File file = new File(dir, fileName);

            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8));
            writer.append(stringBuilder);
            writer.flush();

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    private static StringBuilder parse6Inputs(String input1, String input2, String input3, String input4, String input5, String input6, String resultDirectoryName, Boolean writeFile) {
        StringBuilder sb = new StringBuilder();
        MyDecoder myDecoder = new MyDecoder();

        //find out which MAX-SAT solver produced the output
        String solverOutputName = input1;
        if (solverOutputName.contains("maxino")) {
            maxsatSolverName = "maxino";
        } else if (solverOutputName.contains("open-WBO")) {
            maxsatSolverName = "OpenWBO";
        } else if (solverOutputName.contains("qmaxsat")) {
            maxsatSolverName = "QMaxSAT";
        } else {
            System.err.println("Output from unknown solver -> aborting execution.");
            System.exit(-9);
        }


        //read + parse solver output = .txt-file
        int[] model = myDecoder.parseSolverOutput(input1);
        double costsComputedBySolver = myDecoder.getCostsComputedBySolver();

        //read + parse mapping solver variable to string variable = .csv-file
        SortedMap<Variable, Long> variable2id = new TreeMap<>();
        variable2id = myDecoder.parseMappingVariable2Id(input2);

        //read + parse solver id to solver name = .csv-file
        SortedMap<String, Integer> solver2Id = myDecoder.parseSolver2Id(input3);

        //read + parse instance id to instance name = .csv-file
        SortedMap<String, Integer> instance2Id = myDecoder.parseInstance2Id(input4);

        //read + parse cost-file = .csv-file
        List<List<String>> costFileData = new ArrayList<List<String>>();

        // read cost/time-file
        try (CSVReader csvReader = new CSVReader(new FileReader(input5));) {
            String[] values = null;
            while ((values = csvReader.readNext()) != null) {
                costFileData.add(Arrays.asList(values));
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        //read + parse solve-file = .csv-file
        List<List<String>> solveFileData = new ArrayList<List<String>>();

        // read solve-file
        try (CSVReader csvReader = new CSVReader(new FileReader(input6));) {
            String[] values = null;
            while ((values = csvReader.readNext()) != null) {
                solveFileData.add(Arrays.asList(values));
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        int noOfAllSolvers = costFileData.get(0).size() - 1;    // -1 because first entry is just "instance" and not a solver name
        int noOfAllInstances = costFileData.size() - 1;    // -1 because the first row is not an instance but the name of the solvers

        int[][] allSolved = new int[noOfAllSolvers][noOfAllInstances];
        long[][] allCosts = new long[noOfAllSolvers][noOfAllInstances];

        // start with 1 because first entry is neither an algorithm nor an instance
        for (int i = 1; i <= noOfAllSolvers; i++) {
            for (int j = 1; j <= noOfAllInstances; j++) {
                allSolved[i - 1][j - 1] = Integer.parseInt(solveFileData.get(j).get(i));
                allCosts[i - 1][j - 1] = (long) (Double.parseDouble(costFileData.get(j).get(i)) * 1000);    // convert cut time to milliseconds!! -> weight must be an int greater than 0!
            }
        }

        //create solve and cost matrix for picked solvers and instances
        int noOfPickedAlgorithms = solver2Id.size();
        int noOfPickedInstances = instance2Id.size();
        int[][] solveMatrix = new int[noOfPickedAlgorithms][noOfPickedInstances];
        long[][] costMatrix = new long[noOfPickedAlgorithms][noOfPickedInstances];

        int algIndex = -1;

        for (Map.Entry<String, Integer> entry : solver2Id.entrySet()) {
            algIndex = entry.getValue();
            String solverName = entry.getKey();
            int solverOriginalId = -1;
            for (int i = 1; i < solveFileData.get(0).size(); i++) {     // first entry is not a solver
                if (solveFileData.get(0).get(i).equals(solverName)) {
                    solverOriginalId = i;
                }
            }

            int instanceIndex = -1;
            for (Map.Entry<String, Integer> instanceEntry : instance2Id.entrySet()) {
                instanceIndex = instanceEntry.getValue();
                String instanceName = instanceEntry.getKey();
                int instanceOriginalId = -1;
                for (int i = 1; i < solveFileData.size(); i++) {        // first entry is not an instance
                    if (solveFileData.get(i).get(0).equals(instanceName)) {
                        instanceOriginalId = i;
                    }
                }

                if (solverOriginalId != -1 && instanceOriginalId != -1) {   // index -1 because ids start with 1 not 0
                    solveMatrix[algIndex - 1][instanceIndex - 1] = Integer.parseInt(solveFileData.get(instanceOriginalId).get(solverOriginalId));
                    costMatrix[algIndex - 1][instanceIndex - 1] = (long) (Double.parseDouble(costFileData.get(instanceOriginalId).get(solverOriginalId)) * 1000);    // convert cut time to milliseconds!! -> weight must be an int greater than 0!
                }
            }
        }


        //compute order of algorithms
        List<Integer> totalOrderBySolver = myDecoder.computeTotalOrderOfGivenSolverOutput(model, variable2id);
        //System.out.println(totalOrderBySolver.toString());

        //map ordered algorithms to alg. names
        List<String> totalOrderNames = myDecoder.computeTotalOrderByName(totalOrderBySolver, solver2Id, instance2Id);

        //compute actual costs of this ordering -> would need original input file with costs!
        long costsForPickedInstances = myDecoder.computeTotalOrderCosts(totalOrderBySolver, solveMatrix, costMatrix);


        //compute costs of this ordering for all instances
        long costsForAllInstances = 0;
        boolean[] alreadySolved = new boolean[noOfAllInstances];

        for (int i = 0; i < totalOrderBySolver.size(); i++) {
            String solverName = totalOrderNames.get(i);
            int solverOriginalId = -1;

            for (int j = 1; j < costFileData.get(0).size(); j++) {     // first entry is not a solver
                if (solveFileData.get(0).get(j).equals(solverName)) {
                    solverOriginalId = j;
                }
            }

            if (solverOriginalId != -1) {
                for (int j = 1; j < costFileData.size(); j++) {             // first entry is not an instance
                    if (!alreadySolved[j - 1]) {
                        costsForAllInstances += (long) (Double.parseDouble(costFileData.get(j).get(solverOriginalId)) * 1000);  // conver cut time to millisecond
                        if (allSolved[solverOriginalId - 1][j - 1] == 1) {
                            alreadySolved[j - 1] = true;
                        }
                    }
                }
            }

        }

        //compute number of solvers needed to solve trainingsset
        List<Integer> orderedSolvers = new LinkedList<>();
        orderedSolvers.addAll(totalOrderBySolver);
        List<Integer> unsolvedInstances = new LinkedList<>();

        for (int i = 0; i < noOfPickedInstances; i++) {
            unsolvedInstances.add(i);
        }

        int counterSolverTrainingsset = 0;
        while (!unsolvedInstances.isEmpty() && !orderedSolvers.isEmpty()) {
            //get first solver and remove the now solved instances from the list unsolvedInstances -> done
            int chosenSolver = orderedSolvers.get(0);
            counterSolverTrainingsset++;
            ((LinkedList<Integer>) orderedSolvers).removeFirstOccurrence(chosenSolver);

            for (int i = 0; i < noOfPickedInstances; i++) {
                if (solveMatrix[chosenSolver - 1][i] == 1) {    // -1 because index = id - 1
                    if (unsolvedInstances.contains(i)) {
                        ((LinkedList<Integer>) unsolvedInstances).removeFirstOccurrence(i);
                    }
                }
            }
        }

        //compute number of solvers needed to solve all instances
        List<Integer> orderedSolversAllInstances = new LinkedList<>();
        orderedSolversAllInstances.addAll(totalOrderBySolver);
        //System.out.println(totalOrderBySolver.toString());
        //System.out.println(orderedSolversAllInstances.toString());
        List<Integer> unsolvedInstancesAll = new LinkedList<>();

        for (int i = 0; i < noOfAllInstances; i++) {
            unsolvedInstancesAll.add(i);
        }

        int counterSolverAllInstances = 0;
        while (!unsolvedInstancesAll.isEmpty() && !orderedSolversAllInstances.isEmpty()) {
            //get first solver and remove the now solved instances from the list unsolvedInstances -> done
            int chosenSolver = orderedSolversAllInstances.get(0);
            counterSolverAllInstances++;
            ((LinkedList<Integer>) orderedSolversAllInstances).removeFirstOccurrence(chosenSolver);

            String chosenSolverName = null;
            for (Map.Entry<String, Integer> entry : solver2Id.entrySet()) {
                if (chosenSolver == entry.getValue()) {
                    chosenSolverName = entry.getKey();
                }
            }

            int solverOriginalId = -1;
            for (int j = 1; j < costFileData.get(0).size(); j++) {     // first entry is not a solver
                if (costFileData.get(0).get(j).equals(chosenSolverName)) {
                    solverOriginalId = j;
                }
            }

            for (int i = 0; i < noOfAllInstances; i++) {
                if (allSolved[solverOriginalId - 1][i] == 1) {    // -1 because index = id - 1
                    if (unsolvedInstancesAll.contains(i)) {
                        ((LinkedList<Integer>) unsolvedInstancesAll).removeFirstOccurrence(i);
                    }
                }
            }
        }


        //write output for file

        // instances
        sb.append("Size of trainingsset (number of randomly picked instances) = ").append(noOfPickedInstances).append(System.lineSeparator());
        sb.append("Number of available instances = ").append(noOfAllInstances).append(System.lineSeparator());
        sb.append("Trainingsset:").append(System.lineSeparator());
        int instanceCounter = 0;
        for (String instance : instance2Id.keySet()) {
            instanceCounter++;
            sb.append(instanceCounter).append(") ").append(instance).append(System.lineSeparator());
        }

        // runtime of max-sat solver
        sb.append(System.lineSeparator());
        sb.append("Runtime of MAX-SAT Solver " + maxsatSolverName + " in nanoseconds: " + numberFormat.format(myDecoder.getSolverRuntimeNS())).append(System.lineSeparator());
        sb.append("Runtime of MAX-SAT Solver " + maxsatSolverName + " in seconds: " + myDecoder.getSolverRuntimeS()).append(System.lineSeparator());

        // solvers
        sb.append(System.lineSeparator());
        sb.append("Number of algorithms = ").append(noOfPickedAlgorithms);

        // total order generated by max-sat solver
        sb.append(System.lineSeparator());
        sb.append("Total order of algorithms generated by MAX-SAT Solver " + maxsatSolverName + ":").append(System.lineSeparator());
        for (int i = 0; i < totalOrderNames.size(); i++) {
            sb.append(i + 1).append(". ").append(totalOrderNames.get(i)).append(System.lineSeparator());
        }

        // other format for data-file
        sb.append(System.lineSeparator());
        sb.append("Total order of algorithms generated by MAX-SAT Solver " + maxsatSolverName + ":").append(System.lineSeparator());
        for (int i = 0; i < totalOrderNames.size(); i++) {
            if (i == (totalOrderNames.size() - 1)) {
                sb.append(totalOrderNames.get(i)).append(System.lineSeparator());
            } else {
                sb.append(totalOrderNames.get(i)).append(",");
            }
        }


        // total order with IDs
        sb.append(System.lineSeparator());
        sb.append("Total order of algorithms with ID generated by MAX-SAT Solver " + maxsatSolverName + ":").append(System.lineSeparator());
        for (int i = 0; i < totalOrderBySolver.size(); i++) {
            if (i == (totalOrderBySolver.size() - 1)) {
                sb.append(totalOrderBySolver.get(i)).append(System.lineSeparator());
            } else {
                sb.append(totalOrderBySolver.get(i)).append(",");
            }
        }

        // costs
        sb.append(System.lineSeparator());
        double costsBySolverInSeconds = costsComputedBySolver / 1000;
        double costsInSeconds = ((double) costsForPickedInstances) / 1000;
        double costsForAllInstancesInSeconds = ((double) costsForAllInstances) / 1000;
        sb.append("Costs of this order computed by the MAX-SAT Solver " + maxsatSolverName + " = ").append(numberFormat.format(costsComputedBySolver)).append(System.lineSeparator());
        sb.append("Costs of this order computed by the MAX-SAT Solver " + maxsatSolverName + " (in seconds) = ").append(costsBySolverInSeconds).append(System.lineSeparator());
        sb.append("Actual costs of this order (independently computed) = ").append(costsForPickedInstances).append(System.lineSeparator());
        sb.append("Actual costs of this order (in seconds, independently computed) = ").append(costsInSeconds).append(System.lineSeparator());
        sb.append("Costs of this order for all available instances = ").append(costsForAllInstances).append(System.lineSeparator());
        sb.append("Costs of this order for all available instances (in seconds) = ").append(costsForAllInstancesInSeconds).append(System.lineSeparator());

        // number of solvers needed to solve instances
        sb.append(System.lineSeparator());
        sb.append("Based on the order computed by the MAX-SAT Solver " + maxsatSolverName + ":").append(System.lineSeparator());
        sb.append("Number of algorithms needed to solve trainingsset = ").append(counterSolverTrainingsset).append(System.lineSeparator());
        sb.append("Number of algorithms needed to solve all instances = ").append(counterSolverAllInstances).append(System.lineSeparator());

        // write output file

        String uniqueTime;
        int wIndex = input1.lastIndexOf('W');
        int hyphenIndex = input1.indexOf('-', wIndex);
        int dotIndex = input1.lastIndexOf('.');

        uniqueTime = input1.substring(hyphenIndex + 1, dotIndex);
        String decoderOutputFilename = "decoder-output-" + maxsatSolverName + "-" + uniqueTime + ".txt";

        if (writeFile) {
            writeFile(resultDirectoryName, decoderOutputFilename, sb);
        }

        return sb;
    }

    private static StringBuilder parse7Inputs(String input1, String input2, String input3, String input4, String input5, String input6, String input7, String resultDirectoryName, Boolean writeFile) {
        StringBuilder sb = new StringBuilder();
        MyDecoder myDecoder = new MyDecoder();

        sb = parse6Inputs(input1, input2, input3, input4, input5, input6, resultDirectoryName, false);


        myDecoder.parseGreedyOutput(input7);
        //write greedy output in stringbuilder/to file
        StringBuilder sbWithGreedyOutput = myDecoder.addGreedyOutput(sb);

        String uniqueTime;
        int wIndex = input1.lastIndexOf('W');
        int hyphenIndex = input1.indexOf('-', wIndex);
        int dotIndex = input1.lastIndexOf('.');

        uniqueTime = input1.substring(hyphenIndex + 1, dotIndex);
        String decoderOutputFilename = "decoder-output-" + maxsatSolverName + "-" + uniqueTime + ".txt";

        if (writeFile) {
            writeFile(resultDirectoryName, decoderOutputFilename, sbWithGreedyOutput);
        }

        return sbWithGreedyOutput;
    }

    public static void main(String[] args) {

        numberFormat.setMaximumFractionDigits(340);

        if (args.length == 2) {
            // inputs: 1. Arg = all_time.csv, 2. Arg = all_solved.csv
            //EX.: decoder/src/main/resources/qdimacs-all_time.csv decoder/src/main/resources/qdimacs-all_solved.csv

            String costFileName = args[0];
            String solveFileName = args[1];

            String resourceBaseDirectoryName = "decoder" + File.separator + "src" + File.separator + "main" + File.separator +"resources";
            String solverOutputDirectoryName = "solver-output";
            String encodingVariable2SolverVariableDirectoryName = "encoding_variable-2-solver_variable-mapping";
            String greedyOutputDirectoryName = "greedy-output";
            String instance2idDirectoryName = "instance2id-mapping";
            String solver2idDirectoryName = "solver2id-mapping";

            // automatically run through certain directories and read + filter the remaining necessary files
            // adapt runtype for different runs: 200_instances-20_solvers, 300_instances-20_solvers, 400_instances-20_solvers
            String runType = "200_instances-20_solvers";
            // adapt maxsat solver name: maxino, OpenWBO, QMaxSAT
            String maxSatSolverName = "OpenWBO";

            String solverOutputPath = resourceBaseDirectoryName + File.separator + solverOutputDirectoryName +
                    File.separator + runType + File.separator + maxSatSolverName;
            String encodingV2SolverVPath = resourceBaseDirectoryName + File.separator + encodingVariable2SolverVariableDirectoryName +
                    File.separator + runType;
            String greedyOutputPath = resourceBaseDirectoryName + File.separator + greedyOutputDirectoryName +
                    File.separator + runType;
            String instance2idPath = resourceBaseDirectoryName + File.separator + instance2idDirectoryName +
                    File.separator + runType;
            String solver2idPath = resourceBaseDirectoryName + File.separator + solver2idDirectoryName +
                    File.separator + runType;
            String resultDirectoryName = "decoder" + File.separator + "results" + File.separator + runType + File.separator + maxSatSolverName;
            // testing
            //String resultDirectoryName = "results" + File.separator + "tests";


            // get solver output
            try (Stream<Path> walk = Files.walk(Paths.get(solverOutputPath))) {

                List<String> result = walk.filter(Files::isRegularFile)
                        .map(x -> x.toString()).collect(Collectors.toList());

                for (String res : result) {
                    String solverOutputFilename = res;
                    String encodingV2SolverVFilename = null;
                    String greedyOutputFilename = null;
                    String instance2idFilename = null;
                    String solver2idFilename = null;

                    String uniqueTime;
                    int wIndex = solverOutputFilename.lastIndexOf('W');
                    int hyphenIndex = solverOutputFilename.indexOf('-', wIndex);
                    int dotIndex = solverOutputFilename.lastIndexOf('.');
                    System.out.println(solverOutputFilename);

                    uniqueTime = solverOutputFilename.substring(hyphenIndex + 1, dotIndex);
                    //System.out.println(uniqueTime);

                    // get Wcnf-... file = solver variable to string variable = encoding variable to solver variable
                    try (Stream<Path> walkEncodingV2SolverV = Files.walk(Paths.get(encodingV2SolverVPath))) {

                        List<String> resultEncodingV2SolverV = walkEncodingV2SolverV.map(x -> x.toString())
                                .filter(f -> f.contains(uniqueTime))
                                .collect(Collectors.toList());

                        if (!resultEncodingV2SolverV.isEmpty()) {
                            encodingV2SolverVFilename = resultEncodingV2SolverV.get(0);
                        }

                    } catch (IOException e) {
                        e.printStackTrace();
                    }

                    // get solver2id
                    try (Stream<Path> walkSolver2Id = Files.walk(Paths.get(solver2idPath))) {

                        List<String> resultSolver2Id = walkSolver2Id.map(x -> x.toString())
                                .filter(f -> f.contains(uniqueTime))
                                .collect(Collectors.toList());

                        if (!resultSolver2Id.isEmpty()) {
                            solver2idFilename = resultSolver2Id.get(0);
                        }

                    } catch (IOException e) {
                        e.printStackTrace();
                    }

                    // get instance2id
                    try (Stream<Path> walkInstance2Id = Files.walk(Paths.get(instance2idPath))) {

                        List<String> resultInstance2Id = walkInstance2Id.map(x -> x.toString())
                                .filter(f -> f.contains(uniqueTime))
                                .collect(Collectors.toList());

                        if (!resultInstance2Id.isEmpty()) {
                            instance2idFilename = resultInstance2Id.get(0);
                        }

                    } catch (IOException e) {
                        e.printStackTrace();
                    }

                    // (optionally) get greedy-file
                    try (Stream<Path> walkGreedyOutput = Files.walk(Paths.get(greedyOutputPath))) {

                        List<String> resultGreedyOutput = walkGreedyOutput.map(x -> x.toString())
                                .filter(f -> f.contains(uniqueTime))
                                .collect(Collectors.toList());

                        if (!resultGreedyOutput.isEmpty()) {
                            greedyOutputFilename = resultGreedyOutput.get(0);
                        }

                    } catch (IOException e) {
                        e.printStackTrace();
                    }

                    if (greedyOutputFilename != null) {
                        parse7Inputs(solverOutputFilename, encodingV2SolverVFilename, solver2idFilename, instance2idFilename, costFileName, solveFileName, greedyOutputFilename, resultDirectoryName, true);
                    } else {
                        parse6Inputs(solverOutputFilename, encodingV2SolverVFilename, solver2idFilename, instance2idFilename, costFileName, solveFileName, resultDirectoryName, true);
                    }

                }

            } catch (IOException e) {
                e.printStackTrace();
            }

            System.exit(98);
        } else if (args.length == 6) {
            /* Example:
        decoder/src/main/resources/solver-output/200_instances-20_solvers/maxino/maxino-output-Wcnf-2020-02-19-18-52-13.txt decoder/src/main/resources/encoding_variable-2-solver_variable-mapping/200_instances-20_solvers/Wcnf-2020-02-19-18-52-13.csv decoder/src/main/resources/solver2id-mapping/200_instances-20_solvers/solver2id-2020-02-19-18-52-13.csv decoder/src/main/resources/instance2id-mapping/200_instances-20_solvers/instance2id-2020-02-19-18-52-13.csv decoder/src/main/resources/qdimacs-all_time.csv decoder/src/main/resources/qdimacs-all_solved.csv
         */

            // adapt runtype for different runs: 200_instances-20_solvers, 300_instances-20_solvers, 400_instances-20_solvers
            String runType = "200_instances-20_solvers";
            // adapt maxsat solver name: maxino, OpenWBO, QMaxSAT
            String maxSatSolverName = "maxino";

            String resultDirectoryName = "decoder" + File.separator + "results" + File.separator + runType + File.separator + maxSatSolverName;

            StringBuilder sb = parse6Inputs(args[0], args[1], args[2], args[3], args[4], args[5], resultDirectoryName,true);
            System.exit(99);
        } else if (args.length == 7) {
            /* Examples
        decoder/src/main/resources/solver-output/200_instances-20_solvers/maxino/maxino-output-Wcnf-2020-02-19-18-52-13.txt decoder/src/main/resources/encoding_variable-2-solver_variable-mapping/200_instances-20_solvers/Wcnf-2020-02-19-18-52-13.csv decoder/src/main/resources/solver2id-mapping/200_instances-20_solvers/solver2id-2020-02-19-18-52-13.csv decoder/src/main/resources/instance2id-mapping/200_instances-20_solvers/instance2id-2020-02-19-18-52-13.csv decoder/src/main/resources/qdimacs-all_time.csv decoder/src/main/resources/qdimacs-all_solved.csv decoder/src/main/resources/greedy-output/200_instances-20_solvers/Greedy-2020-02-19-18-52-13.csv
         */

            // adapt runtype for different runs: 200_instances-20_solvers, 300_instances-20_solvers, 400_instances-20_solvers
            String runType = "200_instances-20_solvers";
            // adapt maxsat solver name: maxino, OpenWBO, QMaxSAT
            String maxSatSolverName = "maxino";

            String resultDirectoryName = "decoder" + File.separator + "results" + File.separator + runType + File.separator + maxSatSolverName;

            StringBuilder sb = parse7Inputs(args[0], args[1], args[2], args[3], args[4], args[5], args[6], resultDirectoryName, true);
            System.exit(100);
        }else{
            System.out.println("Usage:\n1. Argument = solver-output\n2. Argument = mapping solver variable to string variable\n" +
                    "3. Argument = solver id to solver name\n4. Argument = instance id to instance name\n" + "5. Argument = cost-file\n" +
                    "6. Argument = solve-file\n" +
                    "optional 7. Argument = greedy-output");
            System.exit(-10);
        }

    }
}
