import com.opencsv.CSVReader;
import org.logicng.formulas.FormulaFactory;
import org.logicng.formulas.Literal;
import org.logicng.formulas.Variable;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.*;

public class MyDecoder {

    private FormulaFactory formulaFactory;

    private double costsComputedBySolver;
    private double solverRuntimeS;
    private double solverRuntimeNS;

    private List<Integer> orderGreedy;
    private int numberNeededSolvers;
    private List<String> orderWithNamesGreedy;
    private double costsGreedy;
    private double costsAllInstancesGreedy;
    private int numberNeededSolversAllInstances;

    public MyDecoder() {
        this.formulaFactory = new FormulaFactory();
    }

    private List<List<String>> readCsvFile(String filename) {
        List<List<String>> mappingFileData = new ArrayList<List<String>>();

        try (CSVReader csvReader = new CSVReader(new FileReader(filename));) {
            String[] values = null;
            while ((values = csvReader.readNext()) != null) {
                mappingFileData.add(Arrays.asList(values));
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return mappingFileData;
    }


    // first + second entry = variable, third = id
    public SortedMap<Variable, Long> parseMappingVariable2Id(String filename) {
        SortedMap<Variable, Long> variable2id = new TreeMap<>();

        List<List<String>> mappingFileData = readCsvFile(filename);

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                Literal v = null;
                long id = 0;

                if (row.size() == 3) {
                    String variableFirstIndex = row.get(0);
                    String secondIndex = row.get(1);
                    id = Long.parseLong(row.get(2));

                    v = this.formulaFactory.literal(variableFirstIndex + "," + secondIndex, true);
                } else if (row.size() == 2) {
                    String variable = row.get(0);
                    id = Long.parseLong(row.get(1));

                    v = this.formulaFactory.literal(variable, true);
                }

                if (v != null) {
                    variable2id.put(v.variable(), id);
                }
            }
        }

        return variable2id;
    }

    public SortedMap<String, Integer> parseSolver2Id(String filename) {
        SortedMap<String, Integer> solver2Id = new TreeMap<>();
        List<List<String>> mappingFileData = readCsvFile(filename);

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                if (row.size() == 2) {
                    String solverName = row.get(0);
                    int id = Integer.parseInt(row.get(1));

                    solver2Id.put(solverName, id);
                }
            }
        }

        return solver2Id;
    }

    public SortedMap<String, Integer> parseInstance2Id(String filename) {
        SortedMap<String, Integer> instance2Id = new TreeMap<>();
        List<List<String>> mappingFileData = readCsvFile(filename);

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                if (row.size() == 2) {
                    String instanceName = row.get(0);
                    int id = Integer.parseInt(row.get(1));

                    instance2Id.put(instanceName, id);
                }
            }
        }

        return instance2Id;
    }

    /*
    1. line: order of solvers in numbers
    2. line: order of solvers by names
    3. line: costs for this order for these instances
    4. line = optional: costs for this order for all instances
     */
    public void parseGreedyOutput(String filename) {
        List<List<String>> mappingFileData = readCsvFile(filename);

        this.orderGreedy = new LinkedList<>();
        this.orderWithNamesGreedy = new LinkedList<>();

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                String firstEntry = row.get(0);

                switch (firstEntry) {
                    case "order":
                        for (int j = 1; j < row.size(); j++) {  // first entry is just "order" and not a solver
                            this.orderGreedy.add(Integer.parseInt(row.get(j)));
                        }
                        break;
                    case "number-needed-solvers":
                        this.numberNeededSolvers = Integer.parseInt(row.get(1));
                        break;
                    case "order-names":
                        for (int j = 1; j < row.size(); j++) {  // first entry is just "order-names" and not a solver
                            this.orderWithNamesGreedy.add(row.get(j));
                        }
                        break;
                    case "costs":
                        this.costsGreedy = Double.parseDouble(row.get(1));
                        break;
                    case "costs-all-instances":
                        this.costsAllInstancesGreedy = Double.parseDouble(row.get(1));
                        break;
                    case "number-needed-solvers-all-instances":
                        this.numberNeededSolversAllInstances = Integer.parseInt(row.get(1));
                        break;
                    default:
                        break;
                }
            }
        }

    }

    public StringBuilder addGreedyOutput(StringBuilder stringBuilder) {

        stringBuilder.append(System.lineSeparator());

        //order
        stringBuilder.append("Total order of algorithms generated by the greedy algorithm:").append(System.lineSeparator());
        for (int i = 0; i < this.orderWithNamesGreedy.size(); i++) {
            stringBuilder.append(i + 1).append(". ").append(this.orderWithNamesGreedy.get(i)).append(System.lineSeparator());
        }

        //costs
        stringBuilder.append(System.lineSeparator());
        stringBuilder.append("Costs of this order computed by the greedy algorithm = ").append(this.costsGreedy).append(System.lineSeparator());
        stringBuilder.append("Costs of this order computed by the greedy algorithm (in seconds) = ").append(this.costsGreedy / 1000).append(System.lineSeparator());
        stringBuilder.append("Costs of this order for all available instances = ").append(this.costsAllInstancesGreedy).append(System.lineSeparator());
        stringBuilder.append("Costs of this order for all available instances (in seconds) = ").append(this.costsAllInstancesGreedy / 1000).append(System.lineSeparator());

        // number of needed solvers
        stringBuilder.append(System.lineSeparator());
        stringBuilder.append("Based on the order computed by the greedy algorithm:").append(System.lineSeparator());
        stringBuilder.append("Number of algorithms needed to solve trainingsset = ").append(this.numberNeededSolvers).append(System.lineSeparator());
        stringBuilder.append("Number of algorithms needed to solve all instances = ").append(this.numberNeededSolversAllInstances).append(System.lineSeparator());


        return stringBuilder;
    }

    public int[] parseSolverOutput(String filename) {
        int[] model = null;

        boolean foundRuntimeS = false;
        boolean foundRuntimeNS = false;
        boolean foundRuntimes = false;

        try {
            BufferedReader reader = new BufferedReader(new FileReader(filename));
            String line = reader.readLine();

	    while (line != null) {
                String[] parts = line.trim().split("\\s+"); // split at whitespace(s)
                if (parts.length >= 2) {
                    String indicator = parts[0];

                    switch (indicator) {
                        case "o":   // optimization solution
                            this.costsComputedBySolver = Double.parseDouble(parts[1]);
                            break;
                        case "v":   // variables
                            model = new int[parts.length - 1];          // -1 because first entry is just "v" and not a variable
                            for (int i = 1; i < parts.length; i++) {    // start at 1 because 0 is just "v" and not a variable
                                if (parts[i].startsWith("-")) {         // negative variable
                                    model[i - 1] = Integer.parseInt(parts[i].substring(1)) * -1;
                                } else {                                // positive variable
                                    model[i - 1] = Integer.parseInt(parts[i]);
                                }
                            }
                            break;
                        case "c":   // comments
                            // filter runtime of solver in seconds and nanoseconds from comments
                            // avoid unnecessary checks after all runtimes have already been parsed
                            if (!foundRuntimes) {
                                if (parts[1].equalsIgnoreCase("runtime") && parts.length == 5) {
                                    // not equals because there might be a ":" included
                                    // => "ns" has to be checked first, bc otherwise case "s" would include case "ns"
                                    if (parts[3].contains("ns")) {
                                        this.solverRuntimeNS = Double.parseDouble(parts[4]);
                                        //System.out.println("Found runtime in ns: " + this.solverRuntimeNS);
                                        foundRuntimeNS = true;
                                        if (foundRuntimeS) {
                                            foundRuntimes = true;
                                        }
                                    } else if (parts[3].contains("s")) { // not equals because there might be a ":" included
                                        this.solverRuntimeS = Double.parseDouble(parts[4]);
                                        //System.out.println("Found runtime in s: " + this.solverRuntimeS);
                                        foundRuntimeS = true;
                                        if (foundRuntimeNS) {
                                            foundRuntimes = true;
                                        }
                                    }
                                }
                            }
                            break;
                        default:
                            break;
                    }
                }


                // read next line
                line = reader.readLine();
            }
            reader.close();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return model;
    }

    public double getCostsComputedBySolver() {
        return costsComputedBySolver;
    }

    public void setCostsComputedBySolver(double costsComputedBySolver) {
        this.costsComputedBySolver = costsComputedBySolver;
    }


    public List<String> computeTotalOrderByName(List<Integer> totOrd, SortedMap<String, Integer> solver2id, SortedMap<String, Integer> instance2id) {

        List<String> totOrderNames = new LinkedList<>();

        for (int i : totOrd) {
            //TODO
            for (Map.Entry<String, Integer> entry : solver2id.entrySet()) {
                if (entry.getValue() == i) {
                    String solverName = entry.getKey();
                    totOrderNames.add(solverName);
                }
            }
        }

        //System.out.println("Total order: " + totOrderNames);
        return totOrderNames;
    }

    public List<Integer> computeTotalOrderOfGivenSolverOutput(int[] model, SortedMap<Variable, Long> variable2id) {
        SortedSet<Literal> solutionLiterals = new TreeSet<>();

        // map id to Variable/Literal
        for (int i : model) {
            int id = Math.abs(i);
            //System.out.println(id);
            Variable v;
            for (Map.Entry<Variable, Long> entry : variable2id.entrySet()) {
                //System.out.println(entry.getValue());
                if (entry.getValue() == id) {
                    v = entry.getKey();
                    String vName = v.name();
                    //System.out.println(vName);
                    boolean phase;
                    if (i < 0) {
                        phase = false;
                    } else {
                        phase = true;
                    }
                    solutionLiterals.add(formulaFactory.literal(vName, phase));
                }
            }
        }

        List<Integer> totalOrder = new Stack<>();

        for (Literal lit : solutionLiterals) {
            //System.out.println(lit.name());
            int firstAlg;
            int secondAlg;
            if (lit.name().startsWith("ord")) {
                String[] indices = lit.name().split(",\\s*");

                if (!lit.phase()) {
                    firstAlg = Integer.parseInt(indices[1]);
                    secondAlg = Integer.parseInt(indices[0].substring(3));
                } else {
                    firstAlg = Integer.parseInt(indices[0].substring(3));
                    secondAlg = Integer.parseInt(indices[1]);
                }

                if (!totalOrder.contains(firstAlg) && !totalOrder.contains(secondAlg)) {
                    totalOrder.add(firstAlg);
                    totalOrder.add(secondAlg);
                    //System.out.println(totalOrder.toString());
                } else if (!totalOrder.contains(firstAlg) && totalOrder.contains(secondAlg)) {
                    totalOrder.add(totalOrder.indexOf(secondAlg), firstAlg);
                    //System.out.println(totalOrder.toString());

                } else if (totalOrder.contains(firstAlg) && !totalOrder.contains(secondAlg)) {
                    totalOrder.add(secondAlg);
                    //System.out.println(totalOrder.toString());


                } else if (totalOrder.contains(firstAlg) && totalOrder.contains(secondAlg)) {
                    if (totalOrder.indexOf(firstAlg) > totalOrder.indexOf(secondAlg)) {
                        int indexToRemove = totalOrder.indexOf(firstAlg);
                        totalOrder.remove(indexToRemove);

                        totalOrder.add(totalOrder.indexOf(secondAlg), firstAlg);
                        //System.out.println(totalOrder.toString());

                    } else {
                        //System.out.println(totalOrder.toString());
                    }
                }

            }
        }
        //System.out.println("Total order: " + totalOrder.toString());
        return totalOrder;
    }

    public long computeTotalOrderCosts(List<Integer> totalOrder, int[][] solveMatrix, long[][] costMatrix) {
        long costs = 0;
        int noOfAlgorithms = solveMatrix.length;
        int noOfInstances = solveMatrix[0].length;
        boolean[] alreadySolved = new boolean[noOfInstances];

        for (Integer i : totalOrder) {  // i corresponds to the number of the algorithm which start at 1 => use i-1 for arrays as they start with index 0
            for (int j = 0; j < noOfInstances; j++) {
                if (!alreadySolved[j]) {
                    costs += costMatrix[i - 1][j];
                    if (solveMatrix[i - 1][j] == 1) {
                        alreadySolved[j] = true;
                    }
                }
            }
        }
        //System.out.println("Actual costs of this order: " + costs);
        return costs;
    }

    public double getSolverRuntimeS() {
        return solverRuntimeS;
    }

    public void setSolverRuntimeS(double solverRuntimeS) {
        this.solverRuntimeS = solverRuntimeS;
    }

    public double getSolverRuntimeNS() {
        return solverRuntimeNS;
    }

    public void setSolverRuntimeNS(double solverRuntimeNS) {
        this.solverRuntimeNS = solverRuntimeNS;
    }
}
