///////////////////////////////////////////////////////////////////////////
//                   __                _      _   ________               //
//                  / /   ____  ____ _(_)____/ | / / ____/               //
//                 / /   / __ \/ __ `/ / ___/  |/ / / __                 //
//                / /___/ /_/ / /_/ / / /__/ /|  / /_/ /                 //
//               /_____/\____/\__, /_/\___/_/ |_/\____/                  //
//                           /____/                                      //
//                                                                       //
//               The Next Generation Logic Library                       //
//                                                                       //
///////////////////////////////////////////////////////////////////////////
//                                                                       //
//  Copyright 2015-20xx Christoph Zengler                                //
//                                                                       //
//  Licensed under the Apache License, Version 2.0 (the "License");      //
//  you may not use this file except in compliance with the License.     //
//  You may obtain a copy of the License at                              //
//                                                                       //
//  http://www.apache.org/licenses/LICENSE-2.0                           //
//                                                                       //
//  Unless required by applicable law or agreed to in writing, software  //
//  distributed under the License is distributed on an "AS IS" BASIS,    //
//  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or      //
//  implied.  See the License for the specific language governing        //
//  permissions and limitations under the License.                       //
//                                                                       //
///////////////////////////////////////////////////////////////////////////

import org.logicng.formulas.FType;
import org.logicng.formulas.Formula;
import org.logicng.formulas.Literal;
import org.logicng.formulas.Variable;
import org.logicng.predicates.CNFPredicate;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.SortedMap;
import java.util.TreeMap;
import java.util.TreeSet;

/**
 * A dimacs file writer for a formula.  Writes the internal data structure of the formula to a weighted dimacs file.
 * @version 1.2
 * @since 1.2
 */
public class MyFormulaWeightedDimacsFileWriter {


    private static SortedMap<Variable, Long> variable2id = new TreeMap<>();


    private static final CNFPredicate CNF_PREDICATE = new CNFPredicate();

    /**
     * Private constructor.
     */
    private MyFormulaWeightedDimacsFileWriter() {
        // Intentionally left empty.
    }

    /**
     * Writes a given formula's internal data structure as a dimacs file.  Must only be called with a formula which is in CNF.
     * @param fileName     the file name of the dimacs file to write
     * @param formulaHardClauses      the formula containing all hard clauses
     * @param formulaSoftClauses    the formula containing all soft clauses
     * @param softClausesWeights    a map of the soft clauses and their weights
     * @param topWeight     the top weight being the sum of all weights + 1
     * @param writeMapping indicates whether an additional file for translating the ids to variable names shall be written
     * @throws IOException              if there was a problem writing the file
     * @throws IllegalArgumentException if the formula was not in CNF
     */
    public static void writeWeightedDimacs(String resultDirectoryName, final String fileName, Formula formulaHardClauses, Formula formulaSoftClauses, Map<Formula, Long> softClausesWeights, long topWeight, boolean writeMapping) throws IOException {
        File dir = new File(resultDirectoryName);
        File file = new File(dir, fileName.endsWith(".cnf") ? fileName : fileName + ".wcnf");
        SortedMap<Variable, Long> var2id = new TreeMap<>();
        long i = 1;
        for (Variable var : new TreeSet<>(formulaHardClauses.variables())) {
            var2id.put(var, i++);
        }
        for (Variable var : new TreeSet<>(formulaSoftClauses.variables())) {
            if (!var2id.containsKey(var)) {
                var2id.put(var, i++);
            }
        }
        if (!formulaHardClauses.holds(CNF_PREDICATE) || !formulaSoftClauses.holds(CNF_PREDICATE)) {
            throw new IllegalArgumentException("Cannot write a non-CNF formula to dimacs.  Convert to CNF first.");
        }
        List<Formula> parts = new ArrayList<>();
        if (formulaHardClauses.type().equals(FType.LITERAL)) {
            parts.add(formulaHardClauses);
        } else {
            for (Formula part : formulaHardClauses) {
                parts.add(part);
            }
        }

        List<Formula> partsSoft = new ArrayList<>();
        if (formulaSoftClauses.type().equals(FType.LITERAL)) {
            partsSoft.add(formulaSoftClauses);
        } else {
            for (Formula partSoft : formulaSoftClauses) {
                partsSoft.add(partSoft);
            }
        }

        // modify for weighted partial max-sat: wcnf and top weight
        StringBuilder sb = new StringBuilder("p wcnf ");
        int partsSize = formulaHardClauses.type().equals(FType.FALSE) ? 1 : parts.size();
        int partsSoftSize = formulaSoftClauses.type().equals(FType.FALSE) ? 1 : partsSoft.size();
        sb.append(var2id.size()).append(" ").append(partsSize + partsSoftSize).append(" ").append(topWeight).append(System.lineSeparator());

        // hard clauses
        for (Formula part : parts) {
            sb.append(topWeight).append(" ");
            for (Literal lit : part.literals()) {
                sb.append(lit.phase() ? "" : "-").append(var2id.get(lit.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        for (Formula softPart : partsSoft) {
            for (Literal literal : softPart.literals()) {
                sb.append(softClausesWeights.get(literal)).append(" ").append(literal.phase() ? "" : "-").append(var2id.get(literal.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        if (formulaHardClauses.type().equals(FType.FALSE) && formulaSoftClauses.type().equals(FType.FALSE)) {
            sb.append(String.format("0%n"));
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
        if (writeMapping) {
            String mappingFileName = (fileName.endsWith(".cnf") ? fileName.substring(0, fileName.length() - 4) : fileName) + ".csv";
            writeMapping(new File(dir, mappingFileName), var2id);
        }

        setVariable2id(var2id);
    }

    /**
     * Writes a given formula's internal data structure as a dimacs file.  Must only be called with a formula which is in CNF.
     * @param fileName     the file name of the dimacs file to write
     * @param formulaHardClauses      the formula containing all hard clauses
     * @param formulaSoftClauses    the formula containing all soft clauses
     * @param softClausesWeights    a map of the soft clauses and their weights
     * @param topWeight     the top weight being the sum of all weights + 1
     * @param writeMapping indicates whether an additional file for translating the ids to variable names shall be written
     * @param costs     the optimal costs of this instance as computed by MyBranchAndBound
     * @throws IOException              if there was a problem writing the file
     * @throws IllegalArgumentException if the formula was not in CNF
     */
    public static void writeWeightedDimacsWithCostInfo(String resultDirectoryName, final String fileName, Formula formulaHardClauses, Formula formulaSoftClauses, Map<Formula, Long> softClausesWeights, long topWeight, boolean writeMapping, long costs) throws IOException {
        File dir = new File(resultDirectoryName);
        File file = new File(dir, fileName.endsWith(".cnf") ? fileName : fileName + ".wcnf");
        SortedMap<Variable, Long> var2id = new TreeMap<>();
        long i = 1;
        for (Variable var : new TreeSet<>(formulaHardClauses.variables())) {
            var2id.put(var, i++);
        }
        for (Variable var : new TreeSet<>(formulaSoftClauses.variables())) {
            if (!var2id.containsKey(var)) {
                var2id.put(var, i++);
            }
        }
        if (!formulaHardClauses.holds(CNF_PREDICATE) || !formulaSoftClauses.holds(CNF_PREDICATE)) {
            throw new IllegalArgumentException("Cannot write a non-CNF formula to dimacs.  Convert to CNF first.");
        }
        List<Formula> parts = new ArrayList<>();
        if (formulaHardClauses.type().equals(FType.LITERAL)) {
            parts.add(formulaHardClauses);
        } else {
            for (Formula part : formulaHardClauses) {
                parts.add(part);
            }
        }

        List<Formula> partsSoft = new ArrayList<>();
        if (formulaSoftClauses.type().equals(FType.LITERAL)) {
            partsSoft.add(formulaSoftClauses);
        } else {
            for (Formula partSoft : formulaSoftClauses) {
                partsSoft.add(partSoft);
            }
        }

        // modify for weighted partial max-sat: wcnf and top weight
        StringBuilder sb = new StringBuilder("c Optimal solution should have costs: " + costs);
        sb.append(System.lineSeparator());
        sb.append("p wcnf ");
        int partsSize = formulaHardClauses.type().equals(FType.FALSE) ? 1 : parts.size();
        int partsSoftSize = formulaSoftClauses.type().equals(FType.FALSE) ? 1 : partsSoft.size();
        sb.append(var2id.size()).append(" ").append(partsSize + partsSoftSize).append(" ").append(topWeight).append(System.lineSeparator());

        // hard clauses
        for (Formula part : parts) {
            sb.append(topWeight).append(" ");
            for (Literal lit : part.literals()) {
                //sb.append(topWeight).append(" ").append(lit.phase() ? "" : "-").append(var2id.get(lit.variable())).append(" ");
                sb.append(lit.phase() ? "" : "-").append(var2id.get(lit.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        for (Formula softPart : partsSoft) {
            for (Literal literal : softPart.literals()) {
                //possible because I have only soft unit clauses
                sb.append(softClausesWeights.get(literal)).append(" ").append(literal.phase() ? "" : "-").append(var2id.get(literal.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        if (formulaHardClauses.type().equals(FType.FALSE) && formulaSoftClauses.type().equals(FType.FALSE)) {
            sb.append(String.format("0%n"));
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
        if (writeMapping) {
            String mappingFileName = (fileName.endsWith(".cnf") ? fileName.substring(0, fileName.length() - 4) : fileName) + ".csv";
            writeMapping(new File(dir, mappingFileName), var2id);
        }

        setVariable2id(var2id);

    }

    /**
     * Writes a given formula's internal data structure as a dimacs file.  Must only be called with a formula which is in CNF.
     * @param fileName     the file name of the dimacs file to write
     * @param formulaHardClauses      the formula containing all hard clauses
     * @param formulaSoftClauses    the formula containing all soft clauses
     * @param softClausesFloatWeights    a map of the soft clauses and their weights in as floating point numbers
     * @param topWeight     the top weight being the sum of all weights + 1
     * @param writeMapping indicates whether an additional file for translating the ids to variable names shall be written
     * @param costs     the optimal costs of this instance as computed by MyBranchAndBound
     * @throws IOException              if there was a problem writing the file
     * @throws IllegalArgumentException if the formula was not in CNF
     */
    public static void writeWeightedDimacsWithCostInfoFloatWeights(String resultDirectoryName, final String fileName, Formula formulaHardClauses, Formula formulaSoftClauses, Map<Formula, Float> softClausesFloatWeights, long topWeight, boolean writeMapping, long costs) throws IOException {
        File dir = new File(resultDirectoryName);
        File file = new File(dir, fileName.endsWith(".cnf") ? fileName : fileName + ".wcnf");
        SortedMap<Variable, Long> var2id = new TreeMap<>();
        long i = 1;
        for (Variable var : new TreeSet<>(formulaHardClauses.variables())) {
            var2id.put(var, i++);
        }
        for (Variable var : new TreeSet<>(formulaSoftClauses.variables())) {
            if (!var2id.containsKey(var)) {
                var2id.put(var, i++);
            }
        }
        if (!formulaHardClauses.holds(CNF_PREDICATE) || !formulaSoftClauses.holds(CNF_PREDICATE)) {
            throw new IllegalArgumentException("Cannot write a non-CNF formula to dimacs.  Convert to CNF first.");
        }
        List<Formula> parts = new ArrayList<>();
        if (formulaHardClauses.type().equals(FType.LITERAL)) {
            parts.add(formulaHardClauses);
        } else {
            for (Formula part : formulaHardClauses) {
                parts.add(part);
            }
        }

        List<Formula> partsSoft = new ArrayList<>();
        if (formulaSoftClauses.type().equals(FType.LITERAL)) {
            partsSoft.add(formulaSoftClauses);
        } else {
            for (Formula partSoft : formulaSoftClauses) {
                partsSoft.add(partSoft);
            }
        }

        // modify for weighted partial max-sat: wcnf and top weight
        StringBuilder sb = new StringBuilder("c Optimal solution should have costs: " + costs);
        sb.append(System.lineSeparator());
        sb.append("p wcnf ");
        int partsSize = formulaHardClauses.type().equals(FType.FALSE) ? 1 : parts.size();
        int partsSoftSize = formulaSoftClauses.type().equals(FType.FALSE) ? 1 : partsSoft.size();
        sb.append(var2id.size()).append(" ").append(partsSize + partsSoftSize).append(" ").append(topWeight).append(System.lineSeparator());

        // hard clauses
        for (Formula part : parts) {
            for (Literal lit : part.literals()) {
                sb.append(topWeight).append(" ").append(lit.phase() ? "" : "-").append(var2id.get(lit.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        for (Formula softPart : partsSoft) {
            for (Literal literal : softPart.literals()) {
                sb.append(softClausesFloatWeights.get(literal)).append(" ").append(literal.phase() ? "" : "-").append(var2id.get(literal.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        if (formulaHardClauses.type().equals(FType.FALSE) && formulaSoftClauses.type().equals(FType.FALSE)) {
            sb.append(String.format("0%n"));
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
        if (writeMapping) {
            String mappingFileName = (fileName.endsWith(".cnf") ? fileName.substring(0, fileName.length() - 4) : fileName) + ".csv";
            writeMapping(new File(dir, mappingFileName), var2id);
        }
        //TODO necessary?
        //setVariable2id(var2id);

    }

    /**
     * Writes a given formula's internal data structure as a dimacs file.  Must only be called with a formula which is in CNF.
     * @param fileName     the file name of the dimacs file to write
     * @param formula      the formula
     * @param writeMapping indicates whether an additional file for translating the ids to variable names shall be written
     * @throws IOException              if there was a problem writing the file
     * @throws IllegalArgumentException if the formula was not in CNF
     */
    public static void writeDimacs(String resultDirectoryName, final String fileName, Formula formula, boolean writeMapping) throws IOException {
        File dir = new File(resultDirectoryName);
        File file = new File(dir, fileName.endsWith(".cnf") ? fileName : fileName + ".cnf");
        SortedMap<Variable, Long> var2id = new TreeMap<>();
        long i = 1;
        for (Variable var : new TreeSet<>(formula.variables())) {
            var2id.put(var, i++);
        }
        if (!formula.holds(CNF_PREDICATE)) {
            throw new IllegalArgumentException("Cannot write a non-CNF formula to dimacs.  Convert to CNF first.");
        }
        List<Formula> parts = new ArrayList<>();
        if (formula.type().equals(FType.LITERAL)) {
            parts.add(formula);
        } else {
            for (Formula part : formula) {
                parts.add(part);
            }
        }
        StringBuilder sb = new StringBuilder("p cnf ");
        int partsSize = formula.type().equals(FType.FALSE) ? 1 : parts.size();
        sb.append(var2id.size()).append(" ").append(partsSize).append(System.lineSeparator());

        for (Formula part : parts) {
            for (Literal lit : part.literals()) {
                sb.append(lit.phase() ? "" : "-").append(var2id.get(lit.variable())).append(" ");
            }
            sb.append(String.format(" 0%n"));
        }
        if (formula.type().equals(FType.FALSE)) {
            sb.append(String.format("0%n"));
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
        if (writeMapping) {
            String mappingFileName = (fileName.endsWith(".cnf") ? fileName.substring(0, fileName.length() - 4) : fileName) + ".csv";
            writeMapping(new File(dir, mappingFileName), var2id);
        }
        setVariable2id(var2id);

    }

    private static void writeMapping(File mappingFile, SortedMap<Variable, Long> var2id) throws IOException {
        StringBuilder sb = new StringBuilder();
        for (Map.Entry<Variable, Long> entry : var2id.entrySet()) {
            sb.append(entry.getKey()).append(",").append(entry.getValue()).append(System.lineSeparator());
        }
        try (BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mappingFile), StandardCharsets.UTF_8))) {
            writer.append(sb);
            writer.flush();
        }
    }

    public static SortedMap<Variable, Long> getVariable2id() {
        return variable2id;
    }

    public static void setVariable2id(SortedMap<Variable, Long> v2id) {
        variable2id = v2id;
    }
}
