import org.logicng.formulas.Formula;
import org.logicng.formulas.FormulaFactory;
import org.logicng.formulas.Literal;
import org.logicng.formulas.Variable;
import org.logicng.io.writers.FormulaDimacsFileWriter;
import org.logicng.solvers.MaxSATSolver;
import org.logicng.solvers.maxsat.algorithms.MaxSAT;

import java.io.IOException;
import java.util.*;

public class MyMaxSatEncoding {

    private String resultDirectoryName;

    private final int noOfAlgorithms;
    private final int noOfInstances;

    private long[][] solveMatrix;
    private long[][] costMatrix;
    private float[][] floatCostMatrix;

    private double expectedNumberOfClauses;

    private FormulaFactory formulaFactory;

    // lists for variables
    private List<Literal> costVariables;
    private List<Literal> solveVariables;
    private List<Literal> helperVariables;
    private List<Literal> orderVariables;
    private List<Literal> negativeOrderVariables;

    // arrays for variables
    private Literal[][] costVariablesPerAlgorithm;
    private Literal[][] solveVariablesPerAlgorithm;
    private Literal[][] helperVariablesPerAlgorithm;
    private List<List<Literal>> orderVariablesPerAlgorithm;

    private Literal[][] orderVariablesNxNArray;

    private Literal[][] helperVariablesPerInstance;

    // lists for clauses
    private List<Formula> solveUnitClauses;
    private List<Formula> atLeastOneHelper;
    private List<Formula> notSolveImpliesNotHelper;
    private List<Formula> helperImpliesCost;
    private List<Formula> atMostOneHelperPerInstance;
    private List<Formula> solveAndOrdImplyNotHelper;
    private List<Formula> solveAndOrdImplyNotCost;
    private List<Formula> ordAndCostImplyCost;
    private List<Formula> orderingTransitivity;

    private List<Formula> softClauses;
    private Map<Formula, Long> softClausesWeights;
    private Map<Formula, Float> softClausesFloatWeights;


    Formula allHardClauses;
    Formula allSoftClauses;

    MyMaxSatEncoding(String resultDirectoryName, int noOfAlgorithms, int noOfInstances) {
        this.resultDirectoryName = resultDirectoryName;

        this.noOfAlgorithms = noOfAlgorithms;
        this.noOfInstances = noOfInstances;

        this.expectedNumberOfClauses = 0;
        // n*m hard unit clauses for s(i,j) or -s(i,j) according to input solve matrix
        this.expectedNumberOfClauses += noOfAlgorithms * noOfInstances;
        // n*(n-1)*(n-2) clauses for linear ordering
        this.expectedNumberOfClauses += (noOfAlgorithms * (noOfAlgorithms - 1) * (noOfAlgorithms - 2));
        // m clauses for at least one helper variable must be true for each instance
        this.expectedNumberOfClauses += noOfInstances;
        // (m*n*(n-1))/2 clauses for at most one helper variables must be true for each instance
        this.expectedNumberOfClauses += ((double) (noOfInstances * noOfAlgorithms * (noOfAlgorithms - 1)) / 2);
        // n*m clauses for -s(i,j) -> -h(i,j)
        this.expectedNumberOfClauses += noOfAlgorithms * noOfInstances;
        // m*n*(n-1) clauses for (s(i,k) and ord(i,j)) -> -h(j,k)
        this.expectedNumberOfClauses += noOfInstances * noOfAlgorithms * (noOfAlgorithms - 1);
        // n*m clauses for h(i,j) -> c(i,j)
        this.expectedNumberOfClauses += noOfAlgorithms * noOfInstances;
        // m*n*(n-1) clauses for (ord(i,j) and c(j,k)) -> c(i,k)
        this.expectedNumberOfClauses += noOfInstances * noOfAlgorithms * (noOfAlgorithms - 1);
        // m*n*(n-1) clauses for (ord(i,j) and s(i,k)) -> -c(j,k)
        this.expectedNumberOfClauses += noOfInstances * noOfAlgorithms * (noOfAlgorithms - 1);
        // n*m soft clauses for -c(i,j)
        this.expectedNumberOfClauses += noOfAlgorithms * noOfInstances;

        this.formulaFactory = new FormulaFactory();

        // lists for variables
        this.costVariables = new LinkedList<>();
        this.solveVariables = new LinkedList<>();
        this.helperVariables = new LinkedList<>();
        this.orderVariables = new LinkedList<>();
        this.negativeOrderVariables = new LinkedList<>();

        // lists for clauses
        this.solveUnitClauses = new LinkedList<>();
        this.atLeastOneHelper = new LinkedList<>();
        this.notSolveImpliesNotHelper = new LinkedList<>();
        this.helperImpliesCost = new LinkedList<>();
        this.atMostOneHelperPerInstance = new LinkedList<>();
        this.solveAndOrdImplyNotHelper = new LinkedList<>();
        this.solveAndOrdImplyNotCost = new LinkedList<>();
        this.ordAndCostImplyCost = new LinkedList<>();
        this.orderingTransitivity = new LinkedList<>();

        this.softClauses = new LinkedList<>();
        this.softClausesWeights = new HashMap<>();
        this.softClausesFloatWeights = new HashMap<>();

        // arrays for variables
        this.costVariablesPerAlgorithm = new Literal[noOfAlgorithms][noOfInstances];
        this.solveVariablesPerAlgorithm = new Literal[noOfAlgorithms][noOfInstances];
        this.helperVariablesPerAlgorithm = new Literal[noOfAlgorithms][noOfInstances];

        this.orderVariablesPerAlgorithm = new ArrayList<>(noOfAlgorithms);
        for (int i = 0; i < noOfAlgorithms; i++) {
            this.orderVariablesPerAlgorithm.add(new LinkedList<>());
        }

        this.orderVariablesNxNArray = new Literal[noOfAlgorithms][noOfAlgorithms];

        this.helperVariablesPerInstance = new Literal[noOfInstances][noOfAlgorithms];

    }

    public void startEncoding() {
        /*
        -> check if all instances can be solved by at least one algorithm, and if not: -> this check is now done in TestMatrixGenerator to solve the problems listed below
            - filter this instance otherwise the formula at the end will be unsatisfiable
            - reduce number of instances by 1 -> problem!! variable is final and everything in the constructor would be false
            - update solveMatrix -> like the old one but delete column of the unsolved instance
         */

        // cost variables, solve variables, helper variables
        for (int i = 1; i <= this.noOfAlgorithms; i++) {
            for (int j = 1; j <= this.noOfInstances; j++) {
                Literal costLiteral = this.formulaFactory.literal("c" + i + "," + j, true);
                Literal solveLiteral = this.formulaFactory.literal("s" + i + "," + j, true);
                Literal helperLiteral = this.formulaFactory.literal("h" + i + "," + j, true);
                this.costVariables.add(costLiteral);
                this.solveVariables.add(solveLiteral);
                this.helperVariables.add(helperLiteral);

                this.costVariablesPerAlgorithm[i - 1][j - 1] = costLiteral;
                this.solveVariablesPerAlgorithm[i - 1][j - 1] = solveLiteral;
                this.helperVariablesPerAlgorithm[i - 1][j - 1] = helperLiteral;

                this.helperVariablesPerInstance[j - 1][i - 1] = helperLiteral;
            }
        }

        // hard unit clauses for solve variables: s(i,j) if solveMatrix[i,j] == 1, -s(i,j) if solveMatrix[i,j] == 0
        if (this.solveMatrix != null) {
            if (this.solveMatrix.length == this.noOfAlgorithms && this.solveMatrix[0].length == this.noOfInstances) {
                for (int i = 0; i < noOfAlgorithms; i++) {
                    for (int j = 0; j < noOfInstances; j++) {
                        Formula solveUnitClause;
                        if (this.solveMatrix[i][j] == 0) {
                            solveUnitClause = this.formulaFactory.clause(this.solveVariablesPerAlgorithm[i][j].negative());
                        } else {
                            solveUnitClause = this.formulaFactory.clause(this.solveVariablesPerAlgorithm[i][j]);
                        }
                        if (solveUnitClause != null) {
                            this.solveUnitClauses.add(solveUnitClause);
                        }
                    }
                }
            }
        }

        // ordering variables
        for (int i = 1; i <= this.noOfAlgorithms; i++) {
            for (int j = i + 1; j <= this.noOfAlgorithms; j++) {
                Literal oV = this.formulaFactory.literal("ord" + i + "," + j, true);
                this.orderVariables.add(oV);
                this.orderVariablesPerAlgorithm.get(i - 1).add(oV);
                this.orderVariablesNxNArray[i - 1][j - 1] = oV;
            }
        }

        // negative ordering variables -> stand for -ord(j,i)
        // -> needed for clauses for transitivity
        for (int i = this.noOfAlgorithms; i > 0; i--) {
            for (int j = i - 1; j > 0; j--) {
                Literal oV = this.formulaFactory.literal("ord" + i + "," + j, true);
                this.negativeOrderVariables.add(oV);
                this.orderVariablesPerAlgorithm.get(i - 1).add(oV);
                this.orderVariablesNxNArray[i - 1][j - 1] = oV;
            }
        }

        // create clauses for at least one helper variable must be true for each instance: V h(i,j) (i = 1 -> n) und 1 <= j <= m
        for (int i = 0; i < this.noOfInstances; i++) {
            Formula atLeastOneHelperTruePerInstance = this.formulaFactory.clause(this.helperVariablesPerInstance[i]);
            this.atLeastOneHelper.add(atLeastOneHelperTruePerInstance);
        }

        // create clauses for -s(i,j) -> -h(i,j)
        // and for h(i,j) -> c(i,j)
        for (int i = 0; i < this.noOfAlgorithms; i++) {
            for (int j = 0; j < this.noOfInstances; j++) {
                // literal.negative returns always(!) a negative literal
                // != literal.negate returns the opposite of the literal
                Literal negativeHelper = this.helperVariablesPerAlgorithm[i][j].negative();
                Formula notSImpliesNotH = this.formulaFactory.clause(this.solveVariablesPerAlgorithm[i][j], negativeHelper);
                this.notSolveImpliesNotHelper.add(notSImpliesNotH);

                Formula hImpliesC = this.formulaFactory.clause(negativeHelper, this.costVariablesPerAlgorithm[i][j]);
                this.helperImpliesCost.add(hImpliesC);
            }
        }

        // create clauses for at most one helper variable must be true for one instance
        // -[h(i,k) and h(j,k)] with i,j = algorithms, k = instance
        for (int i = 0; i < this.noOfInstances; i++) {
            for (int j = 0; j < this.noOfAlgorithms; j++) {
                for (int k = j + 1; k < this.noOfAlgorithms; k++) {
                    Literal firstHelper = this.helperVariablesPerInstance[i][j];
                    Literal secondHelper = this.helperVariablesPerInstance[i][k];

                    Formula atMostOneHelperTruePerInstance = this.formulaFactory.clause(firstHelper.negative(), secondHelper.negative());
                    this.atMostOneHelperPerInstance.add(atMostOneHelperTruePerInstance);
                }
            }
        }

        // a) create clauses for [s(i,k) and ord(i,j)] -> -h(j,k) with i,j = algorithms, k = instance
        // b) + clauses for [s(i,k) and ord(i,j)] -> -c(j,k)
        // c) + clauses for [ord(i,j) and c(j,k)] -> c(i,k)
        for (int i = 0; i < this.noOfAlgorithms; i++) {
            for (int j = 0; j < this.noOfAlgorithms; j++) {
                if (i != j) {   // there are no ord variables ord(i,i)
                    for (int k = 0; k < this.noOfInstances; k++) {
                        Literal solveVariable = this.solveVariablesPerAlgorithm[i][k];
                        Literal helperVariable = this.helperVariablesPerAlgorithm[j][k];
                        Literal costVariable = this.costVariablesPerAlgorithm[j][k];

                        Literal costVariableFirstAlgorithm = this.costVariablesPerAlgorithm[i][k];

                        if (i < j) {
                            Literal ordVariable = this.orderVariablesNxNArray[i][j];
                            // a)
                            Formula solveAndOrdImplyNotHelperVariable = this.formulaFactory.clause(solveVariable.negate(), ordVariable.negate(), helperVariable.negate());
                            this.solveAndOrdImplyNotHelper.add(solveAndOrdImplyNotHelperVariable);
                            // b)
                            Formula solveAndOrdImplyNotCostVariable = this.formulaFactory.clause(solveVariable.negate(), ordVariable.negate(), costVariable.negate());
                            this.solveAndOrdImplyNotCost.add(solveAndOrdImplyNotCostVariable);
                            // c)
                            Formula ordAndCostImplyCostVariable = this.formulaFactory.clause(ordVariable.negate(), costVariable.negate(), costVariableFirstAlgorithm);
                            this.ordAndCostImplyCost.add(ordAndCostImplyCostVariable);

                        } else if (i > j) { // opposite ord variable must be used in clause: ord(i,j) becomes -ord(j,i)
                            Literal ordVariable = this.orderVariablesNxNArray[j][i].negative();
                            // a)
                            Formula solveAndOrdImplyNotHelperVariable = this.formulaFactory.clause(solveVariable.negate(), ordVariable.negate(), helperVariable.negate());
                            this.solveAndOrdImplyNotHelper.add(solveAndOrdImplyNotHelperVariable);
                            // b)
                            Formula solveAndOrdImplyNotCostVariable = this.formulaFactory.clause(solveVariable.negate(), ordVariable.negate(), costVariable.negate());
                            this.solveAndOrdImplyNotCost.add(solveAndOrdImplyNotCostVariable);
                            // c)
                            Formula ordAndCostImplyCostVariable = this.formulaFactory.clause(ordVariable.negate(), costVariable.negate(), costVariableFirstAlgorithm);
                            this.ordAndCostImplyCost.add(ordAndCostImplyCostVariable);
                        }
                    }
                }
            }
        }

        // create clauses for [ord(i,j) and ord(j,l)] -> ord(i,l)
        for (int i = 0; i < this.noOfAlgorithms; i++) {
            for (int j = 0; j < this.orderVariablesPerAlgorithm.get(i).size(); j++) {
                Literal firstVariable = this.orderVariablesPerAlgorithm.get(i).get(j);
                String variableName = firstVariable.name();
                //System.out.println("" + i + variableName);

                String[] algorithms = variableName.split(",\\s*");  // necessary because we do not know how many algorithms there are -> if more than 9, we need two positions for the number
                if (algorithms.length > 0) {
                    algorithms[0] = algorithms[0].substring(3); // position of start of first index = 3 (0 = "o", 1 = "r", 2 = "d", 3 = i)
                }

                int firstIndex = Integer.parseInt(algorithms[0]);
                int secondIndex = Integer.parseInt(algorithms[1]);

                // System.out.println("" + firstAlgorithm + " " + secondAlgorithm);

                // distinguish 2 cases: if i < j then we use the positive variables ord(i,j)
                // if i > j we have to use the negative variable -ord(j,i)
                // now look for the second variable in the clause, where the second index of the first variable must now be the first index
                if (firstIndex < secondIndex) {     // i < j
                    List<Literal> secondVariables = this.orderVariablesPerAlgorithm.get(secondIndex - 1);

                    for (Literal secondVariable : secondVariables) {
                        //Literal secondVariable = secondVariables.get(k);
                        String secondVariableName = secondVariable.name();
                        //System.out.println(secondVariableName);
                        String[] algorithmsSecondVariable = secondVariableName.split(",\\s*");
                        if (algorithmsSecondVariable.length > 0) {
                            algorithmsSecondVariable[0] = algorithmsSecondVariable[0].substring(3); // position of start of first index = 3 (0 = "o", 1 = "r", 2 = "d", 3 = i)
                        }
                        int firstIndexSecondVariable = Integer.parseInt(algorithmsSecondVariable[0]);
                        int secondIndexSecondVariable = Integer.parseInt(algorithmsSecondVariable[1]);

                        // distinguish again the 2 cases i < j and i > j for the indices of this variable
                        if (firstIndexSecondVariable < secondIndexSecondVariable) {     // j < l => i < l
                            // get third variable for clause
                            Literal thirdVariable = this.orderVariablesNxNArray[firstIndex - 1][secondIndexSecondVariable - 1];
                            // System.out.println(firstVariable.name() + " " + secondVariable.name() + " " + thirdVariable.name());

                            // now create clause accordingly: [ord(i,j) and ord(j,l)] -> ord(i,l)
                            Formula transitivity = this.formulaFactory.clause(firstVariable.negate(), secondVariable.negate(), thirdVariable);
                            this.orderingTransitivity.add(transitivity);
                            //System.out.println("added " + transitivity.toString());

                        } else {    // j > l
                            // for the clause: ord(j,l) must become -ord(l,j)
                            Literal actualSecondVariableForClause = this.orderVariablesNxNArray[secondIndexSecondVariable - 1][firstIndexSecondVariable - 1].negative();

                            // distinguish 3 cases for third variable: i = l and i < l and i > l
                            if (firstIndex == secondIndexSecondVariable) {  // i = l
                                // do nothing -> such a clause should not exist as cycles are not allowed for a linear ordering: [ord(i,j) and ord(j,i)] -> ord(i,i)
                                //System.out.println("invalid case");
                            } else if (firstIndex < secondIndexSecondVariable) {   // i < l
                                Literal thirdVariable = this.orderVariablesNxNArray[firstIndex - 1][secondIndexSecondVariable - 1];

                                Formula transitivity = this.formulaFactory.clause(firstVariable.negate(), actualSecondVariableForClause.negate(), thirdVariable);
                                this.orderingTransitivity.add(transitivity);
                                //System.out.println("added " + transitivity.toString());

                            } else {    // i > l => ord(i,l) must become -ord(l,i)
                                Literal thirdVariable = this.orderVariablesNxNArray[secondIndexSecondVariable - 1][firstIndex - 1].negative();

                                Formula transitivity = this.formulaFactory.clause(firstVariable.negate(), actualSecondVariableForClause.negate(), thirdVariable);
                                this.orderingTransitivity.add(transitivity);
                                //System.out.println("added " + transitivity.toString());
                            }
                        }
                    }

                } else {    // i > j => ord(i,j) must become -ord(j,i)
                    Literal actualFirstVariable = this.orderVariablesNxNArray[secondIndex - 1][firstIndex - 1].negative();

                    List<Literal> secondVariables = this.orderVariablesPerAlgorithm.get(secondIndex - 1);

                    for (Literal secondVariable : secondVariables) {
                        //Literal secondVariable = secondVariables.get(k);
                        String secondVariableName = secondVariable.name();
                        //System.out.println(secondVariableName);
                        String[] algorithmsSecondVariable = secondVariableName.split(",\\s*");
                        if (algorithmsSecondVariable.length > 0) {
                            algorithmsSecondVariable[0] = algorithmsSecondVariable[0].substring(3); // position of start of first index = 3 (0 = "o", 1 = "r", 2 = "d", 3 = i)
                        }
                        int firstIndexSecondVariable = Integer.parseInt(algorithmsSecondVariable[0]);
                        int secondIndexSecondVariable = Integer.parseInt(algorithmsSecondVariable[1]);

                        // distinguish again the 2 cases i < j and i > j for the indices of this variable
                        if (firstIndexSecondVariable < secondIndexSecondVariable) {     // j < l
                            // get third variable for clause
                            // distinguish 3 cases: i = l and i > l and i < l
                            if (firstIndex == secondIndexSecondVariable) {      // i = l
                                // invalid case -> do nothing

                            } else if (firstIndex < secondIndexSecondVariable) {  // i < l
                                Literal thirdVariable = this.orderVariablesNxNArray[firstIndex - 1][secondIndexSecondVariable - 1];

                                Formula transitivity = this.formulaFactory.clause(actualFirstVariable.negate(), secondVariable.negate(), thirdVariable);
                                this.orderingTransitivity.add(transitivity);

                            } else {        // i > l
                                Literal thirdVariable = this.orderVariablesNxNArray[secondIndexSecondVariable - 1][firstIndex - 1].negative();

                                Formula transitivity = this.formulaFactory.clause(actualFirstVariable.negate(), secondVariable.negate(), thirdVariable);
                                this.orderingTransitivity.add(transitivity);

                            }

                        } else {    // j > l => i > l
                            // for the clause: ord(j,l) must become -ord(l,j)
                            Literal actualSecondVariableForClause = this.orderVariablesNxNArray[secondIndexSecondVariable - 1][firstIndexSecondVariable - 1].negative();

                            // distinguish 3 cases for third variable: i = l and i < l and i > l
                            if (firstIndex == secondIndexSecondVariable) {  // i = l
                                // do nothing -> such a clause should not exist as cycles are not allowed for a linear ordering: [ord(i,j) and ord(j,i)] -> ord(i,i)
                                //System.out.println("invalid case");
                            } else if (firstIndex < secondIndexSecondVariable) {   // i < l -> should not be possible
                                Literal thirdVariable = this.orderVariablesNxNArray[firstIndex - 1][secondIndexSecondVariable - 1];

                                Formula transitivity = this.formulaFactory.clause(actualFirstVariable.negate(), actualSecondVariableForClause.negate(), thirdVariable);
                                this.orderingTransitivity.add(transitivity);
                                //System.out.println("added " + transitivity.toString());
                                System.err.println("\nentered impossible case\n");

                            } else {    // i > l => ord(i,l) must become -ord(l,i)
                                Literal thirdVariable = this.orderVariablesNxNArray[secondIndexSecondVariable - 1][firstIndex - 1].negative();

                                Formula transitivity = this.formulaFactory.clause(actualFirstVariable.negate(), actualSecondVariableForClause.negate(), thirdVariable);
                                this.orderingTransitivity.add(transitivity);
                                //System.out.println("added " + transitivity.toString());
                            }
                        }
                    }

                }
            }

        }


        // create soft clauses -c(i,j) with i = algorithm, j = instance
        for (int i = 0; i < this.noOfAlgorithms; i++) {
            for (int j = 0; j < this.noOfInstances; j++) {
                Formula soft = this.formulaFactory.clause(this.costVariablesPerAlgorithm[i][j].negative());
                this.softClauses.add(soft);
                this.softClausesWeights.put(soft, this.costMatrix[i][j]);
                this.softClausesFloatWeights.put(soft, this.floatCostMatrix[i][j]);
            }
        }


    }

    public void callWmsu3SatSolver() {
        // create MaxSatSolver with weighted MSU3 algorithm
        MaxSATSolver maxSATSolver = MaxSATSolver.wmsu3();
        maxSATSolver.addHardFormula(this.allHardClauses);
        for (Formula soft : this.softClauses) {
            long weight = this.softClausesWeights.get(soft);
            int weightInt = (int) weight;                       // necessary because built-in solver can only handle int
            maxSATSolver.addSoftFormula(soft, weightInt);
        }

        System.out.println("\nWMSU3: ");
        MaxSAT.MaxSATResult result = maxSATSolver.solve();
        System.out.println(result.toString());
        //System.out.println(maxSATSolver.toString());
        //System.out.println(maxSATSolver.stats());
        System.out.println(maxSATSolver.result());
        System.out.println(maxSATSolver.model());
        printTotalOrder(maxSATSolver);
    }

    public void callIncWBOSatSolver() {
        MaxSATSolver maxSATSolver = MaxSATSolver.incWBO();
        maxSATSolver.addHardFormula(this.allHardClauses);
        for (Formula soft : this.softClauses) {
            long weight = this.softClausesWeights.get(soft);
            int weightInt = (int) weight;                       // necessary because built-in solver can only handle int
            maxSATSolver.addSoftFormula(soft, weightInt);
        }

        System.out.println("\nincWBO: ");
        MaxSAT.MaxSATResult result = maxSATSolver.solve();
        System.out.println(result.toString());
        //System.out.println(maxSATSolver.toString());
        //System.out.println(maxSATSolver.stats());
        System.out.println(maxSATSolver.result());
        System.out.println(maxSATSolver.model());
        printTotalOrder(maxSATSolver);
    }

    public void callLinearSUSatSolver() {
        MaxSATSolver maxSATSolver = MaxSATSolver.linearSU();
        maxSATSolver.addHardFormula(this.allHardClauses);
        for (Formula soft : this.softClauses) {
            long weight = this.softClausesWeights.get(soft);
            int weightInt = (int) weight;                       // necessary because built-in solver can only handle int
            maxSATSolver.addSoftFormula(soft, weightInt);
        }

        System.out.println("\nlinearSU: ");
        MaxSAT.MaxSATResult result = maxSATSolver.solve();
        System.out.println(result.toString());
        //System.out.println(maxSATSolver.toString());
        //System.out.println(maxSATSolver.stats());
        System.out.println(maxSATSolver.result());
        System.out.println(maxSATSolver.model());
        printTotalOrder(maxSATSolver);
    }

    public void callWBOSatSolver() {
        MaxSATSolver maxSATSolver = MaxSATSolver.wbo();
        maxSATSolver.addHardFormula(this.allHardClauses);
        for (Formula soft : this.softClauses) {
            long weight = this.softClausesWeights.get(soft);
            int weightInt = (int) weight;                       // necessary because built-in solver can only handle int
            maxSATSolver.addSoftFormula(soft, weightInt);
        }

        System.out.println("\nWBO: ");
        MaxSAT.MaxSATResult result = maxSATSolver.solve();
        System.out.println(result.toString());
        //System.out.println(maxSATSolver.toString());
        //System.out.println(maxSATSolver.stats());
        System.out.println(maxSATSolver.result());
        System.out.println(maxSATSolver.model());
        printTotalOrder(maxSATSolver);
    }

    /* linear US does not support weighted MaxSAT instances currently (28.04.2019)
    public void callLinearUSSatSolver() {
        MaxSATSolver maxSATSolver = MaxSATSolver.linearUS();
        maxSATSolver.addHardFormula(this.allHardClauses);
        for (Formula soft : this.softClauses) {
            int weight = this.softClausesWeights.get(soft);
            maxSATSolver.addSoftFormula(soft, weight);
        }

        System.out.println("linearUS: ");
        MaxSAT.MaxSATResult result = maxSATSolver.solve();
        System.out.println(result.toString());
        System.out.println(maxSATSolver.toString());
        System.out.println(maxSATSolver.stats());
        System.out.println(maxSATSolver.result());
        System.out.println(maxSATSolver.model());
        printTotalOrder(maxSATSolver);
    }
    */

    public void printTotalOrderByName(List<Integer> totOrd) {
        //TestMatrixGenerator testGen = new TestMatrixGenerator();
        SortedMap<String, Integer> solver2id = TestMatrixGenerator.getSolver2Id();
        SortedMap<String, Integer> instance2id = TestMatrixGenerator.getInstance2Id();

        List<String> totOrderNames = new LinkedList<>();

        for (int i : totOrd) {
            for (Map.Entry<String, Integer> entry : solver2id.entrySet()) {
                if (entry.getValue() == i) {
                    String solverName = entry.getKey();
                    totOrderNames.add(solverName);
                }
            }
        }

        System.out.println("Total order: " + totOrderNames);

    }

    public void printTotalOrderOfGivenInput(int[] model, SortedMap<Variable, Long> variable2id) {
        SortedSet<Literal> solutionLiterals = new TreeSet<>();

        // map id to Variable/Literal
        for (int i : model) {
            int id = Math.abs(i);
            //System.out.println(id);
            Variable v;
            for (Map.Entry<Variable, Long> entry : variable2id.entrySet()) {
                //System.out.println(entry.getValue());
                if (entry.getValue() == id) {
                    v = entry.getKey();
                    String vName = v.name();
                    //System.out.println(vName);
                    boolean phase;
                    if (i < 0) {
                        phase = false;
                    } else {
                        phase = true;
                    }
                    solutionLiterals.add(formulaFactory.literal(vName, phase));
                }
            }
        }

        List<Integer> totalOrder = new Stack<>();

        for (Literal lit : solutionLiterals) {
            int firstAlg;
            int secondAlg;
            if (lit.name().startsWith("ord")) {
                String[] indices = lit.name().split(",\\s*");

                if (!lit.phase()) {
                    firstAlg = Integer.parseInt(indices[1]);
                    secondAlg = Integer.parseInt(indices[0].substring(3));
                } else {
                    firstAlg = Integer.parseInt(indices[0].substring(3));
                    secondAlg = Integer.parseInt(indices[1]);
                }

                if (!totalOrder.contains(firstAlg) && !totalOrder.contains(secondAlg)) {
                    totalOrder.add(firstAlg);
                    totalOrder.add(secondAlg);
                    //System.out.println(totalOrder.toString());
                } else if (!totalOrder.contains(firstAlg) && totalOrder.contains(secondAlg)) {
                    totalOrder.add(totalOrder.indexOf(secondAlg), firstAlg);
                    //System.out.println(totalOrder.toString());

                } else if (totalOrder.contains(firstAlg) && !totalOrder.contains(secondAlg)) {
                    totalOrder.add(secondAlg);
                    //System.out.println(totalOrder.toString());


                } else if (totalOrder.contains(firstAlg) && totalOrder.contains(secondAlg)) {
                    if (totalOrder.indexOf(firstAlg) > totalOrder.indexOf(secondAlg)) {
                        int indexToRemove = totalOrder.indexOf(firstAlg);
                        totalOrder.remove(indexToRemove);

                        totalOrder.add(totalOrder.indexOf(secondAlg), firstAlg);
                        //System.out.println(totalOrder.toString());

                    } else {
                        // do nothing -> they are already correctly ordered
                        //System.out.println(totalOrder.toString());
                    }
                }

            }
        }
        System.out.println("Total order: " + totalOrder.toString());
        printTotalOrderByName(totalOrder);
        computeTotalOrderCosts(totalOrder);
    }

    private void printTotalOrder(MaxSATSolver maxSATSolver) {
        //System.out.println(maxSATSolver.model().positiveLiterals());
        //System.out.println(maxSATSolver.model().negativeLiterals());

        if (maxSATSolver.model() != null && maxSATSolver.model().literals() != null) {
            SortedSet<Literal> solutionLiterals = maxSATSolver.model().literals();
            List<Integer> totalOrder = new Stack<>();

            for (Literal lit : solutionLiterals) {
                int firstAlg;
                int secondAlg;
                if (lit.name().startsWith("ord")) {
                    String[] indices = lit.name().split(",\\s*");

                    if (!lit.phase()) {
                        firstAlg = Integer.parseInt(indices[1]);
                        secondAlg = Integer.parseInt(indices[0].substring(3));
                    } else {
                        firstAlg = Integer.parseInt(indices[0].substring(3));
                        secondAlg = Integer.parseInt(indices[1]);
                    }

                    if (!totalOrder.contains(firstAlg) && !totalOrder.contains(secondAlg)) {
                        totalOrder.add(firstAlg);
                        totalOrder.add(secondAlg);
                        //System.out.println(totalOrder.toString());
                    } else if (!totalOrder.contains(firstAlg) && totalOrder.contains(secondAlg)) {
                        totalOrder.add(totalOrder.indexOf(secondAlg), firstAlg);
                        //System.out.println(totalOrder.toString());

                    } else if (totalOrder.contains(firstAlg) && !totalOrder.contains(secondAlg)) {
                        totalOrder.add(secondAlg);
                        //System.out.println(totalOrder.toString());


                    } else if (totalOrder.contains(firstAlg) && totalOrder.contains(secondAlg)) {
                        if (totalOrder.indexOf(firstAlg) > totalOrder.indexOf(secondAlg)) {
                            int indexToRemove = totalOrder.indexOf(firstAlg);
                            totalOrder.remove(indexToRemove);

                            totalOrder.add(totalOrder.indexOf(secondAlg), firstAlg);
                            //System.out.println(totalOrder.toString());

                        } else {
                            // do nothing -> they are already correctly ordered
                            //System.out.println(totalOrder.toString());
                        }
                    }

                }
            }
            System.out.println("Total order: " + totalOrder.toString());
            computeTotalOrderCosts(totalOrder);
        }


    }

    private void computeTotalOrderCosts(List<Integer> totalOrder) {
        long costs = 0;
        boolean[][] alreadySolved = new boolean[this.noOfAlgorithms][this.noOfInstances];

        for (Integer i : totalOrder) {  // i corresponds to the number of the algorithm which start at 1 => use i-1 for arrays as they start with index 0
            for (int j = 0; j < this.noOfInstances; j++) {
                if (!alreadySolved[i - 1][j]) {
                    costs += this.costMatrix[i - 1][j];
                    if (this.solveMatrix[i - 1][j] == 1) {
                        for (int k = 0; k < this.noOfAlgorithms; k++) {
                            alreadySolved[k][j] = true;
                        }
                    }
                }
            }
        }
        System.out.println("Actual costs of this order: " + costs);
    }


    private long computeTopWeight() {
        // initialize with 1 to make sure top is a weight always greater than the sum of the weights of violated soft clauses in the solution
        // which is the sum of all costs in the worst case
        long topWeight = 1;

        if (this.costMatrix != null && this.costMatrix.length == this.noOfAlgorithms && this.costMatrix[0].length == this.noOfInstances) {
            for (int i = 0; i < this.noOfAlgorithms; i++) {
                for (int j = 0; j < this.noOfInstances; j++) {
                    topWeight += this.costMatrix[i][j];
                }
            }
        }
        return topWeight;
    }

    private float computeTopFloatWeight() {
        // initialize with 1 to make sure top is a weight always greater than the sum of the weights of violated soft clauses in the solution
        // which is the sum of all costs in the worst case
        float topWeight = 1;

        if (this.floatCostMatrix != null && this.floatCostMatrix.length == this.noOfAlgorithms && this.floatCostMatrix[0].length == this.noOfInstances) {
            for (int i = 0; i < this.noOfAlgorithms; i++) {
                for (int j = 0; j < this.noOfInstances; j++) {
                    topWeight += this.floatCostMatrix[i][j];
                }
            }
        }
        return topWeight;
    }

    public void allClausesWriteCNF() {
        Formula allSolveUnitClauses = this.formulaFactory.and(this.solveUnitClauses);

        Formula allAtLeastOneHelper = this.formulaFactory.and(this.atLeastOneHelper);
        //System.out.println(this.atLeastOneHelper.size());
        //System.out.println(allAtLeastOneHelper.numberOfOperands() + " " + allAtLeastOneHelper.type());
        Formula allNotSolveImpliesNotHelper = this.formulaFactory.and(this.notSolveImpliesNotHelper);
        //System.out.println(this.notSolveImpliesNotHelper.size());
        //System.out.println(allNotSolveImpliesNotHelper.numberOfOperands() + " " + allNotSolveImpliesNotHelper.type());
        Formula allHelperImpliesCost = this.formulaFactory.and(this.helperImpliesCost);
        //System.out.println(this.helperImpliesCost.size());
        //System.out.println(allHelperImpliesCost.numberOfOperands() + " " + allHelperImpliesCost.type());
        Formula allOrderingTransitivity = this.formulaFactory.and(this.orderingTransitivity);
        //System.out.println(this.orderingTransitivity.size());
        //System.out.println(allOrderingTransitivity.numberOfOperands() + " " + allOrderingTransitivity.type());
        Formula allAtMostOneHelperPerInstance = this.formulaFactory.and(this.atMostOneHelperPerInstance);
        //System.out.println(this.atMostOneHelperPerInstance.size());
        //System.out.println(allAtMostOneHelperPerInstance.numberOfOperands() + " " + allAtMostOneHelperPerInstance.type());
        Formula allSolveAndOrdImplyNotHelper = this.formulaFactory.and(this.solveAndOrdImplyNotHelper);
        //System.out.println(this.solveAndOrdImplyNotHelper.size());
        //System.out.println(allSolveAndOrdImplyNotHelper.numberOfOperands() + " " + allSolveAndOrdImplyNotHelper.type());
        Formula allSolveAndOrdImplyNotCost = this.formulaFactory.and(this.solveAndOrdImplyNotCost);
        //System.out.println(this.solveAndOrdImplyNotCost.size());
        //System.out.println(allSolveAndOrdImplyNotCost.numberOfOperands() + " " + allSolveAndOrdImplyNotCost.type());
        Formula allOrdAndCostImplyCost = this.formulaFactory.and(this.ordAndCostImplyCost);
        //System.out.println(this.ordAndCostImplyCost.size());
        //System.out.println(allOrdAndCostImplyCost.numberOfOperands() + " " + allOrdAndCostImplyCost.type());

        this.allHardClauses = this.formulaFactory.and(allSolveUnitClauses, allAtLeastOneHelper, allNotSolveImpliesNotHelper,
                allHelperImpliesCost, allOrderingTransitivity, allAtMostOneHelperPerInstance, allSolveAndOrdImplyNotHelper,
                allSolveAndOrdImplyNotCost, allOrdAndCostImplyCost);

        this.allSoftClauses = this.formulaFactory.and(this.softClauses);

        System.out.println("Number of operands of hard clauses: " + this.allHardClauses.numberOfOperands());
        System.out.println("Number of operands of soft clauses: " + this.allSoftClauses.numberOfOperands());
        System.out.println("Total number of operands: " + (this.allHardClauses.numberOfOperands() + this.allSoftClauses.numberOfOperands()));

        long top = computeTopWeight();


        try {
            MyFormulaWeightedDimacsFileWriter.writeWeightedDimacs(this.resultDirectoryName,"testWeightedCNF", this.allHardClauses, this.allSoftClauses, this.softClausesWeights, top, true);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void allClausesWriteCNFWithCosts(long costs, String wcnfFileName) {
        Formula allSolveUnitClauses = this.formulaFactory.and(this.solveUnitClauses);

        Formula allAtLeastOneHelper = this.formulaFactory.and(this.atLeastOneHelper);
        //System.out.println(this.atLeastOneHelper.size());
        //System.out.println(allAtLeastOneHelper.numberOfOperands() + " " + allAtLeastOneHelper.type());
        Formula allNotSolveImpliesNotHelper = this.formulaFactory.and(this.notSolveImpliesNotHelper);
        //System.out.println(this.notSolveImpliesNotHelper.size());
        //System.out.println(allNotSolveImpliesNotHelper.numberOfOperands() + " " + allNotSolveImpliesNotHelper.type());
        Formula allHelperImpliesCost = this.formulaFactory.and(this.helperImpliesCost);
        //System.out.println(this.helperImpliesCost.size());
        //System.out.println(allHelperImpliesCost.numberOfOperands() + " " + allHelperImpliesCost.type());
        Formula allOrderingTransitivity = this.formulaFactory.and(this.orderingTransitivity);
        //System.out.println(this.orderingTransitivity.size());
        //System.out.println(allOrderingTransitivity.numberOfOperands() + " " + allOrderingTransitivity.type());
        Formula allAtMostOneHelperPerInstance = this.formulaFactory.and(this.atMostOneHelperPerInstance);
        //System.out.println(this.atMostOneHelperPerInstance.size());
        //System.out.println(allAtMostOneHelperPerInstance.numberOfOperands() + " " + allAtMostOneHelperPerInstance.type());
        Formula allSolveAndOrdImplyNotHelper = this.formulaFactory.and(this.solveAndOrdImplyNotHelper);
        //System.out.println(this.solveAndOrdImplyNotHelper.size());
        //System.out.println(allSolveAndOrdImplyNotHelper.numberOfOperands() + " " + allSolveAndOrdImplyNotHelper.type());
        Formula allSolveAndOrdImplyNotCost = this.formulaFactory.and(this.solveAndOrdImplyNotCost);
        //System.out.println(this.solveAndOrdImplyNotCost.size());
        //System.out.println(allSolveAndOrdImplyNotCost.numberOfOperands() + " " + allSolveAndOrdImplyNotCost.type());
        Formula allOrdAndCostImplyCost = this.formulaFactory.and(this.ordAndCostImplyCost);
        //System.out.println(this.ordAndCostImplyCost.size());
        //System.out.println(allOrdAndCostImplyCost.numberOfOperands() + " " + allOrdAndCostImplyCost.type());

        this.allHardClauses = this.formulaFactory.and(allSolveUnitClauses, allAtLeastOneHelper, allNotSolveImpliesNotHelper,
                allHelperImpliesCost, allOrderingTransitivity, allAtMostOneHelperPerInstance, allSolveAndOrdImplyNotHelper,
                allSolveAndOrdImplyNotCost, allOrdAndCostImplyCost);

        this.allSoftClauses = this.formulaFactory.and(this.softClauses);

        System.out.println("Number of operands of hard clauses: " + allHardClauses.numberOfOperands());
        System.out.println("Number of operands of soft clauses: " + allSoftClauses.numberOfOperands());
        System.out.println("Total number of operands: " + (allHardClauses.numberOfOperands() + allSoftClauses.numberOfOperands()));

        long top = computeTopWeight();

        try {
            MyFormulaWeightedDimacsFileWriter.writeWeightedDimacsWithCostInfo(this.resultDirectoryName, wcnfFileName, this.allHardClauses, this.allSoftClauses, this.softClausesWeights, top, true, costs);
        } catch (IOException e) {
            e.printStackTrace();
        }

        float floatTop = computeTopFloatWeight();

        try {
            MyFormulaWeightedDimacsFileWriter.writeWeightedDimacsWithCostInfoFloatWeights(this.resultDirectoryName,"Float-" + wcnfFileName, this.allHardClauses, this.allSoftClauses, this.softClausesFloatWeights, top, false, costs);
        } catch (IOException e) {
            e.printStackTrace();
        }
    }

    public void hardClausesCNF() {
        Formula allSolveUnitClauses = this.formulaFactory.and(this.solveUnitClauses);
        Formula allAtLeastOneHelper = this.formulaFactory.and(this.atLeastOneHelper);
        Formula allNotSolveImpliesNotHelper = this.formulaFactory.and(this.notSolveImpliesNotHelper);
        Formula allHelperImpliesCost = this.formulaFactory.and(this.helperImpliesCost);
        Formula allOrderingTransitivity = this.formulaFactory.and(this.orderingTransitivity);
        Formula allAtMostOneHelperPerInstance = this.formulaFactory.and(this.atMostOneHelperPerInstance);
        Formula allSolveAndOrdImplyNotHelper = this.formulaFactory.and(this.solveAndOrdImplyNotHelper);
        Formula allSolveAndOrdImplyNotCost = this.formulaFactory.and(this.solveAndOrdImplyNotCost);
        Formula allOrdAndCostImplyCost = this.formulaFactory.and(this.ordAndCostImplyCost);

        this.allHardClauses = this.formulaFactory.and(allSolveUnitClauses, allAtLeastOneHelper, allNotSolveImpliesNotHelper,
                allHelperImpliesCost, allOrderingTransitivity, allAtMostOneHelperPerInstance, allSolveAndOrdImplyNotHelper,
                allSolveAndOrdImplyNotCost, allOrdAndCostImplyCost);

        System.out.println("Number of operands: " + allHardClauses.numberOfOperands());
        System.out.println("Number of atoms: " + allHardClauses.numberOfAtoms());
        System.out.println("Number of nodes: " + allHardClauses.numberOfNodes());


        try {
            FormulaDimacsFileWriter.write("testCNF", allHardClauses, true);
        } catch (IOException e) {
            e.printStackTrace();
        }

    }

    public long[][] getSolveMatrix() {
        return solveMatrix;
    }

    public void setSolveMatrix(long[][] solveMatrix) {
        this.solveMatrix = solveMatrix;
    }

    public long[][] getCostMatrix() {
        return costMatrix;
    }

    public void setCostMatrix(long[][] costMatrix) {
        this.costMatrix = costMatrix;
    }

    public float[][] getFloatCostMatrix() {
        return floatCostMatrix;
    }

    public void setFloatCostMatrix(float[][] floatCostMatrix) {
        this.floatCostMatrix = floatCostMatrix;
    }
}
