/*
    * class to parse input files as generated by class "Test"
    * necessary to run MyBranchAndBoundMoreOptimized and possibly other classes for an already existing testset of instances
    * functionality partly similar to my decoder
 */

import com.opencsv.CSVReader;
import org.logicng.formulas.FormulaFactory;
import org.logicng.formulas.Literal;
import org.logicng.formulas.Variable;

import java.io.*;
import java.util.*;

public class MyParser implements Runnable{

    private static final String resultDirectoryName = "encoder" + File.separator + "results";
    private String uniqueTime;

    private FormulaFactory formulaFactory;

    private String variableMappingFile;
    private String solverMappingFile;
    private String instanceMappingFile;
    private String costFile;
    private String solveFile;

    private SortedMap<Variable, Long> variable2id;
    private SortedMap<String, Integer> solver2Id;
    private SortedMap<String, Integer> instance2Id;

    MyParser(String variableMappingFile, String solverMappingFile, String instanceMappingFile, String costFile, String solveFile) {
        // cut "Wcnf-" in the front and ".csv" in the back
        // substring(): startIndex = inclusive, endIndex = exclusive
        // String.indexOf() returns index of FIRST occurrence
        int wIndex = variableMappingFile.indexOf('W');
        int hyphenIndex = variableMappingFile.indexOf('-', wIndex);
        int dotIndex = variableMappingFile.lastIndexOf('.');
        //System.out.println(variableMappingFile);

        this.uniqueTime = variableMappingFile.substring(hyphenIndex + 1, dotIndex);
        //System.out.println(this.uniqueTime);

        this.variableMappingFile = variableMappingFile;
        this.solverMappingFile = solverMappingFile;
        this.instanceMappingFile = instanceMappingFile;
        this.costFile = costFile;
        this.solveFile = solveFile;

        this.variable2id = new TreeMap<>();
        this.solver2Id = new TreeMap<>();
        this.instance2Id = new TreeMap<>();

        this.formulaFactory = new FormulaFactory();
    }

    private List<List<String>> readCsvFile(String filename) {
        List<List<String>> mappingFileData = new ArrayList<List<String>>();

        try (CSVReader csvReader = new CSVReader(new FileReader(filename));) {
            String[] values = null;
            while ((values = csvReader.readNext()) != null) {
                mappingFileData.add(Arrays.asList(values));
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        return mappingFileData;
    }

    // first + second entry = variable, third = id
    private void parseMappingVariable2Id(String filename) {
        List<List<String>> mappingFileData = readCsvFile(filename);

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                Literal v = null;
                long id = 0;

                if (row.size() == 3) {
                    String variableFirstIndex = row.get(0);
                    String secondIndex = row.get(1);
                    id = Long.parseLong(row.get(2));

                    v = this.formulaFactory.literal(variableFirstIndex + "," + secondIndex, true);
                } else if (row.size() == 2) {
                    String variable = row.get(0);
                    id = Long.parseLong(row.get(1));

                    v = this.formulaFactory.literal(variable, true);
                }

                if (v != null) {
                    this.variable2id.put(v.variable(), id);
                }
            }
        }

    }

    private void parseSolver2Id(String filename) {
        List<List<String>> mappingFileData = readCsvFile(filename);

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                if (row.size() == 2) {
                    String solverName = row.get(0);
                    int id = Integer.parseInt(row.get(1));

                    this.solver2Id.put(solverName, id);
                }
            }
        }

    }

    private void parseInstance2Id(String filename) {
        List<List<String>> mappingFileData = readCsvFile(filename);

        if (mappingFileData != null && !mappingFileData.isEmpty()) {
            for (int i = 0; i < mappingFileData.size(); i++) {
                List<String> row = mappingFileData.get(i);
                if (row.size() == 2) {
                    String instanceName = row.get(0);
                    int id = Integer.parseInt(row.get(1));

                    this.instance2Id.put(instanceName, id);
                }
            }
        }

    }

    @Override
    public void run() {

        parseMappingVariable2Id(this.variableMappingFile);
        parseSolver2Id(this.solverMappingFile);
        parseInstance2Id(this.instanceMappingFile);

        //read + parse cost-file = .csv-file
        List<List<String>> costFileData = new ArrayList<List<String>>();

        // read cost/time-file
        try (CSVReader csvReader = new CSVReader(new FileReader(this.costFile));) {
            String[] values = null;
            while ((values = csvReader.readNext()) != null) {
                costFileData.add(Arrays.asList(values));
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        //read + parse solve-file = .csv-file
        List<List<String>> solveFileData = new ArrayList<List<String>>();

        // read solve-file
        try (CSVReader csvReader = new CSVReader(new FileReader(this.solveFile));) {
            String[] values = null;
            while ((values = csvReader.readNext()) != null) {
                solveFileData.add(Arrays.asList(values));
            }
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        int noOfAllSolvers = costFileData.get(0).size() - 1;    // -1 because first entry is just "instance" and not a solver name
        int noOfAllInstances = costFileData.size() - 1;    // -1 because the first row is not an instance but the name of the solvers

        int[][] allSolved = new int[noOfAllSolvers][noOfAllInstances];
        long[][] allCosts = new long[noOfAllSolvers][noOfAllInstances];

        // start with 1 because first entry is neither an algorithm nor an instance
        for (int i = 1; i <= noOfAllSolvers; i++) {
            for (int j = 1; j <= noOfAllInstances; j++) {
                allSolved[i - 1][j - 1] = Integer.parseInt(solveFileData.get(j).get(i));
                allCosts[i - 1][j - 1] = (long) (Double.parseDouble(costFileData.get(j).get(i)) * 1000);    // convert cut time to milliseconds!! -> weight must be an int greater than 0!
            }
        }

        //create solve and cost matrix for picked solvers and instances
        int noOfPickedAlgorithms = solver2Id.size();
        int noOfPickedInstances = instance2Id.size();

        long[][] solveMatrix = new long[noOfPickedAlgorithms][noOfPickedInstances];
        long[][] costMatrix = new long[noOfPickedAlgorithms][noOfPickedInstances];

        int algIndex = -1;

        for (Map.Entry<String, Integer> entry : solver2Id.entrySet()) {
            algIndex = entry.getValue();
            String solverName = entry.getKey();
            int solverOriginalId = -1;
            for (int i = 1; i < solveFileData.get(0).size(); i++) {     // first entry is not a solver
                if (solveFileData.get(0).get(i).equals(solverName)) {
                    solverOriginalId = i;
                }
            }

            int instanceIndex = -1;
            for (Map.Entry<String, Integer> instanceEntry : instance2Id.entrySet()) {
                instanceIndex = instanceEntry.getValue();
                String instanceName = instanceEntry.getKey();
                int instanceOriginalId = -1;
                for (int i = 1; i < solveFileData.size(); i++) {        // first entry is not an instance
                    if (solveFileData.get(i).get(0).equals(instanceName)) {
                        instanceOriginalId = i;
                    }
                }

                if (solverOriginalId != -1 && instanceOriginalId != -1) {   // index -1 because ids start with 1 not 0
                    solveMatrix[algIndex - 1][instanceIndex - 1] = Long.parseLong(solveFileData.get(instanceOriginalId).get(solverOriginalId));
                    costMatrix[algIndex - 1][instanceIndex - 1] = (long) (Double.parseDouble(costFileData.get(instanceOriginalId).get(solverOriginalId)) * 1000);    // convert cut time to milliseconds!! -> weight must be an int greater than 0!
                }
            }
        }

        // get solver names
        String[] solverNames = new String[solveFileData.get(0).size() - 1];
        for (int i = 1; i < solveFileData.get(0).size(); i++) {
            solverNames[i - 1] = solveFileData.get(0).get(i);
            //System.out.println(solverNames[i - 1]);
        }

        // get instance names
        String[] instanceNames = new String[solveFileData.size() - 1];
        for (int i = 1; i < solveFileData.size(); i++) {
            instanceNames[i - 1] = solveFileData.get(i).get(0);

        }

        System.out.println("start B&B");
        MyBranchAndBoundOptimized myBranchAndBoundOptimized = new MyBranchAndBoundOptimized(resultDirectoryName, uniqueTime, noOfPickedAlgorithms, noOfPickedInstances);
        myBranchAndBoundOptimized.setSolved(solveMatrix);
        myBranchAndBoundOptimized.setCosts(costMatrix);
        myBranchAndBoundOptimized.run();
        System.out.println("end B&B");

        System.out.println("start B&B-PLAIN");
        MyBranchAndBound myBranchAndBound = new MyBranchAndBound(resultDirectoryName, uniqueTime, noOfPickedAlgorithms, noOfPickedInstances);
        myBranchAndBound.setSolved(solveMatrix);
        myBranchAndBound.setCosts(costMatrix);
        myBranchAndBound.run();
        System.out.println("end B&B-PLAIN");

        MyGreedyPortfolioScheduling myGreedyPortfolioScheduling = new MyGreedyPortfolioScheduling(resultDirectoryName, uniqueTime);
        myGreedyPortfolioScheduling.setNoOfPickedSolvers(noOfPickedAlgorithms);
        myGreedyPortfolioScheduling.setNoOfPickedInstances(noOfPickedInstances);
        myGreedyPortfolioScheduling.setSolved(solveMatrix);
        myGreedyPortfolioScheduling.setCosts(costMatrix);
        myGreedyPortfolioScheduling.setDataForAllInstances(solverNames, instanceNames, costFileData, solveFileData);
        myGreedyPortfolioScheduling.setSolver2id(this.solver2Id);
        myGreedyPortfolioScheduling.setInstance2id(this.instance2Id);
        System.out.println("start Greedy");
        myGreedyPortfolioScheduling.startFromParser();
        System.out.println("end Greedy");

    }

}
