import com.opencsv.CSVReader;
import org.logicng.formulas.Variable;
import org.logicng.solvers.MaxSATSolver;
import org.sat4j.maxsat.SolverFactory;
import org.sat4j.maxsat.WeightedMaxSatDecorator;
import org.sat4j.maxsat.reader.WDimacsReader;
import org.sat4j.pb.IPBSolver;
import org.sat4j.reader.DimacsReader;
import org.sat4j.reader.ParseFormatException;
import org.sat4j.reader.Reader;
import org.sat4j.specs.*;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.sql.SQLOutput;
import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.concurrent.TimeUnit;

public class Test {

    private static final String DATE_FORMATTER= "yyyy-MM-dd-HH-mm-ss";

    private static final String resultDirectoryName = "encoder" + File.separator + "results";


    public static void main(String[] args) {

        DateTimeFormatter formatter = DateTimeFormatter.ofPattern(DATE_FORMATTER);
        String uniqueTime = LocalDateTime.now().format(formatter).toString();
        String wcnfFileName = "Wcnf-" + uniqueTime;

        String encodingFileName = "Encoding-" + uniqueTime + ".csv";
        StringBuilder stringBuilder = new StringBuilder();

        if (args.length == 0) {     // run with randomly generated instances

            // for thesis - example data
//            TestMatrixGenerator testMatrixGenerator2 = new TestMatrixGenerator("");
//            testMatrixGenerator2.setTestInstance2();
//
//            MyBranchAndBoundOptimized myBranchAndBoundOptimized2 = new MyBranchAndBoundOptimized(resultDirectoryName, uniqueTime, 4, 5);
//            myBranchAndBoundOptimized2.setSolved(testMatrixGenerator2.getSolveMatrix());
//            myBranchAndBoundOptimized2.setCosts(testMatrixGenerator2.getCostMatrix());
//            myBranchAndBoundOptimized2.run();
//
//            System.exit(99);

            int noOfAlgorithms = 4;
            int noOfInstances = 10;
            MyMaxSatEncoding myMaxSatEncoding = new MyMaxSatEncoding(resultDirectoryName, noOfAlgorithms, noOfInstances);

            // generates random test data, but does not tell the optimal solution
            TestMatrixGenerator testMatrixGenerator = new TestMatrixGenerator(resultDirectoryName);
            testMatrixGenerator.generateMatrices(noOfAlgorithms, noOfInstances);

            //testMatrixGenerator.setTestInstance1(); // 4 algorithms, 5 instances

            long[][] solved = testMatrixGenerator.getSolveMatrix();
            long[][] costs = testMatrixGenerator.getCostMatrix();
            float[][] floatCosts = testMatrixGenerator.getFloatCostMatrix();


            MyBranchAndBound myBranchAndBound = new MyBranchAndBound(resultDirectoryName, uniqueTime, noOfAlgorithms, noOfInstances);
            myBranchAndBound.setSolved(solved);
            myBranchAndBound.setCosts(costs);
            myBranchAndBound.run();

            long bAndBCosts = myBranchAndBound.getUpperBound();

            MyBranchAndBoundOptimized myBranchAndBoundOptimized = new MyBranchAndBoundOptimized(resultDirectoryName, uniqueTime, noOfAlgorithms, noOfInstances);
            myBranchAndBoundOptimized.setSolved(solved);
            myBranchAndBoundOptimized.setCosts(costs);
            myBranchAndBoundOptimized.run();

            //Greedy-Algorithm
            MyGreedyPortfolioScheduling myGreedyPortfolioScheduling = new MyGreedyPortfolioScheduling(resultDirectoryName, uniqueTime);
            myGreedyPortfolioScheduling.setNoOfPickedSolvers(noOfAlgorithms);
            myGreedyPortfolioScheduling.setNoOfPickedInstances(noOfInstances);
            myGreedyPortfolioScheduling.setSolved(solved);
            myGreedyPortfolioScheduling.setCosts(costs);
            myGreedyPortfolioScheduling.setRandomInput(true);
            //myGreedyPortfolioScheduling.setDataForAllInstances(solverNames, instanceNames, costFileData, solveFileData);
            myGreedyPortfolioScheduling.run();

            myMaxSatEncoding.setSolveMatrix(solved);
            myMaxSatEncoding.setCostMatrix(costs);
            myMaxSatEncoding.setFloatCostMatrix(floatCosts);

            System.out.println("start encoding");
            long startTimeEncoding = System.nanoTime();
            myMaxSatEncoding.startEncoding();
            long endTimeEncoding = System.nanoTime();
            System.out.println("end encoding");

            double elapsedTimeEncoding = (double) endTimeEncoding - startTimeEncoding;
            double elapsedTimeInSecondsEncoding = elapsedTimeEncoding * 0.000000001;
            System.out.println("Encoding time in nanoseconds: " + elapsedTimeEncoding);
            System.out.println("Encoding time in seconds: " + elapsedTimeInSecondsEncoding);

            System.out.println("start wcnf");
            myMaxSatEncoding.allClausesWriteCNFWithCosts(bAndBCosts, wcnfFileName);
            System.out.println("end wcnf");

            stringBuilder.append("runtime-in-nanoseconds").append(",").append(elapsedTimeEncoding).append(System.lineSeparator());
            stringBuilder.append("runtime-in-seconds").append(",").append(elapsedTimeInSecondsEncoding).append(System.lineSeparator());

            writeFile(encodingFileName, stringBuilder);

            // call weighted max-sat-solvers

            myMaxSatEncoding.callLinearSUSatSolver();
            myMaxSatEncoding.callWmsu3SatSolver();
            myMaxSatEncoding.callWBOSatSolver();
            myMaxSatEncoding.callIncWBOSatSolver();


        } else if (args.length == 2) { // 1st = file with costs, 2nd = file with solve-info
            //EX: encoder/src/main/resources/qdimacs-all_time.csv encoder/src/main/resources/qdimacs-all_solved.csv

            int noOfPickedInstances = 200;
            int noOfPickedSolvers = 20; // 20 = max for qdimacs-all_solved/time

            List<List<String>> costFileData = new ArrayList<List<String>>();
            List<List<String>> solveFileData = new ArrayList<List<String>>();

            // read cost/time-file
            try (CSVReader csvReader = new CSVReader(new FileReader(args[0]));) {
                String[] values = null;
                while ((values = csvReader.readNext()) != null) {
                    costFileData.add(Arrays.asList(values));
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }

            // read solve-file
            try (CSVReader csvReader = new CSVReader(new FileReader(args[1]));) {
                String[] values = null;
                while ((values = csvReader.readNext()) != null) {
                    solveFileData.add(Arrays.asList(values));
                }
            } catch (FileNotFoundException e) {
                e.printStackTrace();
            } catch (IOException e) {
                e.printStackTrace();
            }

            //System.out.println(costFileData.get(0).get(1));
            //System.out.println(solveFileData.get(1).get(0));

            // get solver names
            String[] solverNames = new String[solveFileData.get(0).size() - 1];
            for (int i = 1; i < solveFileData.get(0).size(); i++) {
                solverNames[i - 1] = solveFileData.get(0).get(i);
                //System.out.println(solverNames[i - 1]);
            }

            // get instance names
            String[] instanceNames = new String[solveFileData.size() - 1];
            for (int i = 1; i < solveFileData.size(); i++) {
                instanceNames[i - 1] = solveFileData.get(i).get(0);

            }

            MyMaxSatEncoding myMaxSatEncoding = new MyMaxSatEncoding(resultDirectoryName, noOfPickedSolvers, noOfPickedInstances);

            // generates random test data, but does not tell the optimal solution
            TestMatrixGenerator testMatrixGenerator = new TestMatrixGenerator(resultDirectoryName);
            testMatrixGenerator.pickRandomlyFromGivenInput(noOfPickedSolvers, noOfPickedInstances, costFileData, solveFileData, uniqueTime);


            long[][] solved = testMatrixGenerator.getSolveMatrix();
            long[][] costs = testMatrixGenerator.getCostMatrix();
            float[][] floatCosts = testMatrixGenerator.getFloatCostMatrix();

            //Branch-and-Bound
            MyBranchAndBound myBranchAndBound = new MyBranchAndBound(resultDirectoryName, uniqueTime, noOfPickedSolvers, noOfPickedInstances);
            myBranchAndBound.setSolved(solved);
            myBranchAndBound.setCosts(costs);
            System.out.println("start B&B-PLAIN");
            myBranchAndBound.run();
            System.out.println("end B&B-PLAIN");

            long bAndBCosts = myBranchAndBound.getUpperBound();

            //Branch-and-Bound optimized
            System.out.println("start B&B");
            MyBranchAndBoundOptimized myBranchAndBoundOptimized = new MyBranchAndBoundOptimized(resultDirectoryName, uniqueTime, noOfPickedSolvers, noOfPickedInstances);
            myBranchAndBoundOptimized.setSolved(solved);
            myBranchAndBoundOptimized.setCosts(costs);
            myBranchAndBoundOptimized.run();
            System.out.println("end B&B");

            //Greedy-Algorithm
            MyGreedyPortfolioScheduling myGreedyPortfolioScheduling = new MyGreedyPortfolioScheduling(resultDirectoryName, uniqueTime);
            myGreedyPortfolioScheduling.setNoOfPickedSolvers(noOfPickedSolvers);
            myGreedyPortfolioScheduling.setNoOfPickedInstances(noOfPickedInstances);
            myGreedyPortfolioScheduling.setSolved(solved);
            myGreedyPortfolioScheduling.setCosts(costs);
            myGreedyPortfolioScheduling.setDataForAllInstances(solverNames, instanceNames, costFileData, solveFileData);
            System.out.println("start Greedy");
            myGreedyPortfolioScheduling.run();
            System.out.println("end Greedy");

            // Encoding
            myMaxSatEncoding.setSolveMatrix(solved);
            myMaxSatEncoding.setCostMatrix(costs);
            myMaxSatEncoding.setFloatCostMatrix(floatCosts);
            System.out.println("start encoding");
            long startTimeEncoding = System.nanoTime();
            myMaxSatEncoding.startEncoding();
            long endTimeEncoding = System.nanoTime();
            System.out.println("end encoding");

            double elapsedTimeEncoding = (double) endTimeEncoding - startTimeEncoding;
            double elapsedTimeInSecondsEncoding = elapsedTimeEncoding * 0.000000001;
            System.out.println("Encoding time in nanoseconds: " + elapsedTimeEncoding);
            System.out.println("Encoding time in seconds: " + elapsedTimeInSecondsEncoding);

            System.out.println("start wcnf");
            myMaxSatEncoding.allClausesWriteCNFWithCosts(bAndBCosts, wcnfFileName);
            System.out.println("end wcnf");

            SortedMap<Variable, Long> variable2id = MyFormulaWeightedDimacsFileWriter.getVariable2id();

            stringBuilder.append("runtime-in-nanoseconds").append(",").append(elapsedTimeEncoding).append(System.lineSeparator());
            stringBuilder.append("runtime-in-seconds").append(",").append(elapsedTimeInSecondsEncoding).append(System.lineSeparator());

            writeFile(encodingFileName, stringBuilder);


        } else if (args.length == 5) { // for running certain algorithms again for already picked and executed data
            //1.Argument = mapping solver variable to string variable
            //2. Argument = solver id to solver name
            //3. Argument = instance id to instance name
            //4. Argument = cost-file
            //5. Argument = solve-file

            //EX.: encoder/results/Wcnf-2020-09-30-22-18-08.csv encoder/results/solver2id-2020-09-30-22-18-08.csv encoder/results/instance2id-2020-09-30-22-18-08.csv encoder/src/main/resources/qdimacs-all_time.csv encoder/src/main/resources/qdimacs-all_solved.csv

            MyParser myParser = new MyParser(args[0], args[1], args[2], args[3], args[4]);
            System.out.println("before parsing");
            myParser.run();
        }
    }

    private static void writeFile(String fileName, StringBuilder stringBuilder) {
        try {
            File dir = new File(resultDirectoryName);
            File file = new File(dir, fileName);

            BufferedWriter writer = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file), StandardCharsets.UTF_8));
            writer.append(stringBuilder);
            writer.flush();

        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }
    }
}
